from __future__ import absolute_import

import functools
import os
import typing

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko import rhosp as rhosp_topology
from tobiko.rhosp import containers as rhosp_containers


CONF = config.CONF
LOG = log.getLogger(__name__)


class ContainerRuntimeFixture(tobiko.SharedFixture):

    runtime: typing.Optional[rhosp_containers.ContainerRuntime] = None

    def setup_fixture(self):
        self.runtime = self.get_runtime()

    def cleanup_fixture(self):
        self.runtime = None

    @staticmethod
    def get_runtime() -> typing.Optional[rhosp_containers.ContainerRuntime]:
        """return handle to the container runtime"""
        return rhosp_containers.PODMAN_RUNTIME


def get_container_runtime() -> rhosp_containers.ContainerRuntime:
    runtime = tobiko.setup_fixture(ContainerRuntimeFixture).runtime
    return runtime


def get_container_runtime_name() -> str:
    return get_container_runtime().runtime_name


def is_podman() -> bool:
    return True


def has_container_runtime() -> bool:
    return True


@functools.lru_cache()
def list_node_containers(ssh_client):
    """returns a list of containers and their run state"""
    return get_container_runtime().list_containers(ssh_client=ssh_client)


def get_container_client(ssh_client=None):
    """returns a list of containers and their run state"""
    return get_container_runtime().get_client(ssh_client=ssh_client)


def list_containers_td(group=None):
    actual_containers_list = list_containers(group)
    return tobiko.TableData(
        get_container_states_list(actual_containers_list),
        columns=['container_host', 'container_name', 'container_state'])


def list_containers(group=None):
    """get list of containers in running state
    from specified node group
    returns : a list of overcloud_node's running containers"""

    # moved here from topology
    # reason : Workaround for :
    # AttributeError: module 'tobiko.openstack.topology' has no
    # attribute 'container_runtime'

    if group is None:
        group = 'compute'
    containers_list = tobiko.Selection()
    openstack_nodes = topology.list_openstack_nodes(group=group)

    for node in openstack_nodes:
        LOG.debug(f"List containers for node {node.name}")
        node_containers_list = list_node_containers(ssh_client=node.ssh_client)
        containers_list.extend(node_containers_list)
    return containers_list


def _select_container(
        container_itr: typing.Union[str, typing.Iterable[str]],
        full_name,
        node_name,
        containers_list_td,
) -> typing.Tuple[typing.Optional[str], tobiko.TableData, list]:
    """Return the status of a single container if present"""
    _name, _attrs, _fail = None, tobiko.TableData(), []
    if isinstance(container_itr, str):
        container_itr = [container_itr]

    # Get selected container attrs tabledata.
    for _name in container_itr:
        if full_name:
            _attrs = containers_list_td.query(
                'container_name == "{}"'.format(_name))
        else:
            container_attrs_rows = []
            for row in containers_list_td:
                if _name in row['container_name']:
                    container_attrs_rows.append(row)
            _attrs = tobiko.TableData(container_attrs_rows)
        if _attrs.empty:
            _fail.append(
                'expected container {} not found on node {} ! : \n\n'.
                format(_name, node_name))
        else:
            break

    return _name, _attrs, _fail


def assert_containers_running(
        group: topology.OpenstackGroupNamesType,
        expected_containers: typing.Iterable[
            typing.Union[str, typing.Iterable[str]]],
        full_name: bool = True,
        bool_check: bool = False,
        nodenames: typing.Iterable[str] = None,
) -> typing.Union[None, bool]:
    """assert that all containers specified in the list are running
    on the specified openstack group(controller or compute etc..)
    if bool_check is True then return only True or false without failing"""

    failures = []

    openstack_nodes = topology.list_openstack_nodes(group=group,
                                                    hostnames=nodenames)
    for node in openstack_nodes:
        node_containers = list_node_containers(ssh_client=node.ssh_client)
        containers_list_td = tobiko.TableData(
            get_container_states_list(node_containers),
            columns=['container_host', 'container_name', 'container_state'])
        # check that the containers are present
        LOG.info('node: {} containers list : {}'.format(
            node.name, containers_list_td.to_string()))
        for container_itr in expected_containers:
            container_name, container_attrs, container_fail = (
                _select_container(container_itr, full_name, node.name,
                                  containers_list_td))

            # check if any container exists
            LOG.info('checking container(s): {}'.format(container_itr))
            if not container_name:
                failures += container_fail
                continue
            else:
                LOG.info('selected container: {}'.format(container_name))

            # If one container exists, check it is running and discard the
            # other options. Only one running container is expected.
            container_running_attrs = container_attrs.query(
                'container_state=="running"')
            if container_running_attrs.empty:
                failures.append(
                    'expected container {} is not running on node {} , '
                    'its state is {}! : \n\n'.format(
                        container_name, node.name,
                        container_attrs['container_state'].values.item()))
            elif len(container_running_attrs) > 1:
                failures.append(
                    'only one running container {} was expected on '
                    'node {}, but got {}! : \n\n'.format(
                        container_name, node.name,
                        len(container_running_attrs)))

    if not bool_check and failures:
        tobiko.fail(
            'container states mismatched:\n{}'.format('\n'.join(failures)),
            rhosp_containers.ContainerMismatchException)

    elif bool_check and failures:
        return False

    else:
        LOG.info('All specified containers are in running state! ')
        return True


def assert_ovn_containers_running():
    if not neutron.has_ovn():
        LOG.info("Networking OVN not configured")
        return
    ovn_containers = [('ovn_metadata_agent', 'ovn_agent'),
                      'ovn_controller']
    potential_groups = ['edpm-compute', 'edpm-networker']
    groups = [group for group in potential_groups if
              group in topology.get_openstack_topology().groups]
    for group in groups:
        assert_containers_running(group, ovn_containers)
    LOG.info("Networking OVN containers verified in running state")


def get_container_states_list(containers_list,
                              include_container_objects=False):
    container_states_list = tobiko.Selection()
    container_states_list.extend([comparable_container_keys(
        container, include_container_objects=include_container_objects) for
                                  container in containers_list])
    return container_states_list


def save_containers_state_to_file(expected_containers_list,):
    expected_containers_td = tobiko.TableData(
        get_container_states_list(expected_containers_list),
        columns=['container_host', 'container_name', 'container_state'])
    expected_containers_td.to_csv(
        rhosp_containers.expected_containers_file)
    return rhosp_containers.expected_containers_file


def comparable_container_keys(container, include_container_objects=False):
    """returns the tuple : 'container_host','container_name',
    'container_state, container object if specified'
     """
    host_or_ip = container.client.base_url.netloc.rsplit('_')[1]
    nodes_matching = topology.list_openstack_nodes(hostnames=[host_or_ip])
    nodename = (nodes_matching[0].name
                if nodes_matching
                else rhosp_topology.ip_to_hostname(host_or_ip))

    # Differenciate between podman_ver3 with podman-py from earlier api
    if include_container_objects:
        return (nodename,
                container.attrs['Names'][0], container.attrs['State'],
                container)
    else:
        return (nodename,
                container.attrs['Names'][0],
                container.attrs['State'])


@functools.lru_cache()
def list_containers_objects_td():
    containers_list = list_containers()
    containers_objects_list_td = tobiko.TableData(
        get_container_states_list(
            containers_list, include_container_objects=True),
        columns=['container_host', 'container_name',
                 'container_state', 'container_object'])
    return containers_objects_list_td


def get_edpm_container(container_name=None, container_host=None,
                       partial_container_name=None):
    """gets an container object by name on specified host
    container"""
    con_obj_td = list_containers_objects_td()
    if partial_container_name and container_host:
        filtered_rows = []
        for row in con_obj_td:
            if partial_container_name in row['container_name']:
                filtered_rows.append(row)
        con_obj_td = tobiko.TableData(filtered_rows)

        container_obj = con_obj_td.query(
            'container_host == "{container_host}"'.format(
                container_host=container_host))
        if not container_obj.empty:
            return container_obj.data[0]['container_object']
    elif container_host:
        container_obj = con_obj_td.query(
            'container_name == "{container_name}"'
            ' and container_host == "{container_host}"'.
            format(container_host=container_host,
                   container_name=container_name))
        if not container_obj.empty:
            return container_obj.data[0]['container_object']
    else:
        container_obj = con_obj_td.query(
            'container_name == "{container_name}"'.
            format(container_name=container_name))
        if not container_obj.empty:
            return container_obj.data[0]['container_object']

    tobiko.fail('container {} not found!'.format(container_name))


def action_on_container(action: str,
                        container_name=None,
                        container_host=None,
                        partial_container_name=None):
    """take a container and perform an action on it
    actions are as defined in : podman/libs/containers.py:14/164"""

    LOG.debug(f"Executing '{action}' action on container "
              f"'{container_name}' "
              f"on host '{container_host}'")

    container_object = get_edpm_container(
        container_name=container_name,
        container_host=container_host,
        partial_container_name=partial_container_name)

    try:
        container_action = getattr(container_object, action)
        container_action()
        LOG.debug(f"Successfully executed '{action}' action on container "
                  f"'{container_name}' on host '{container_host}'")
    except Exception as e:
        LOG.error(f"Error occurred while executing '{action}' action on "
                  f"container '{container_name}' on host '{container_host}': "
                  f"{e}")
        raise


def assert_equal_containers_state(expected_containers_list=None,
                                  timeout=120, interval=2,
                                  recreate_expected=False):

    """
    compare the current containers with the expected containers list
    the container states list is either built from the expected_containers_list
    or is the current containers state compared to the previously created file
    """

    expected_containers_td = None
    if recreate_expected or (
            not expected_containers_list and
            not os.path.exists(rhosp_containers.expected_containers_file)):
        save_containers_state_to_file(list_containers())
        return
    elif expected_containers_list:
        expected_containers_td = tobiko.TableData(
            get_container_states_list(expected_containers_list),
            columns=['container_host', 'container_name', 'container_state'])

    elif os.path.exists(rhosp_containers.expected_containers_file):
        with open(rhosp_containers.expected_containers_file, 'r') as f:
            expected_containers_td = tobiko.TableData.read_csv(f, header=0)

    LOG.info("Comparing current containers with expected containers")

    for attempt in tobiko.retry(timeout=timeout, interval=interval):
        actual_containers_td = list_containers_td()

        LOG.info(f'expected_containers_td: {expected_containers_td}')
        LOG.info(f'actual_containers_td: {actual_containers_td}')

        # execute a `tabledata` diff between the expected and actual containers
        diff_tb = rhosp_containers.tabledata_difference(
            expected_containers_td,
            actual_containers_td)

        if diff_tb.empty:
            LOG.info("All containers are on the same state")
            return

        if attempt.is_last:
            tobiko.fail("Containers state does not match expected state:\n"
                        f"{diff_tb}")



# Copyright (c) 2024 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import functools
import os

from oslo_concurrency import lockutils
from oslo_log import log
from oslo_utils import reflection
from oslo_utils import timeutils

from tobiko.common import _fixture


LOG = log.getLogger(__name__)


def interworker_synched(name):
    """Re-definition of oslo_concurrency.lockutils.synchronized.

    Tobiko needs to re-difine this decorator in order to avoid intra-
    process/worker locks. This is because tobiko is executed in multiple
    processes (using pytest), but each process only runs one thread.

    The intra-process lock should not be applied in tobiko because some of the
    locked methods could be called recurrently.
    Example:
    The creation (setup_fixture) of CirrosPeerServerStackFixture depends on the
    creation of CirrosServerStackFixture, which is also its parent class.
    With intra-process locks, the creation of CirrosServerStackFixture could
    not be started (would be locked by the creation of
    CirrosPeerServerStackFixture).

    The lock name is automatically extended with the fully qualified class
    name of the instance (including module path) to prevent different fixture
    classes from blocking each other.
    """

    def wrap(f):

        @functools.wraps(f)
        def inner(*args, **kwargs):
            # Include the fully qualified class name in the lock to ensure
            # different fixture classes don't block each other
            instance = args[0] if args else None
            if instance is not None:
                class_name = _fixture.get_object_name(instance)
                lock_name = f"{name}:{class_name}"
            else:
                lock_name = name

            t1 = timeutils.now()
            t2 = None
            gotten = True
            f_name = reflection.get_callable_name(f)
            try:
                with lock(lock_name):
                    t2 = timeutils.now()
                    LOG.debug('Lock "%(name)s" acquired by "%(function)s" :: '
                              'waited %(wait_secs)0.3fs',
                              {'name': lock_name,
                               'function': f_name,
                               'wait_secs': (t2 - t1)})
                    return f(*args, **kwargs)
            except lockutils.AcquireLockFailedException:
                gotten = False
            finally:
                t3 = timeutils.now()
                if t2 is None:
                    held_secs = "N/A"
                else:
                    held_secs = "%0.3fs" % (t3 - t2)
                LOG.debug('Lock "%(name)s" "%(gotten)s" by "%(function)s" ::'
                          ' held %(held_secs)s',
                          {'name': lock_name,
                           'gotten': 'released' if gotten else 'unacquired',
                           'function': f_name,
                           'held_secs': held_secs})
        return inner

    return wrap


@contextlib.contextmanager
def lock(name):
    """Re-definition of oslo_concurrency.lockutils.lock that does not apply
    intra-worker locks. Only inter-worker locks are applied.
    """
    from tobiko import config
    lock_path = os.path.expanduser(config.CONF.tobiko.common.lock_dir)

    # Ensure lock directory exists before trying to acquire locks
    os.makedirs(lock_path, exist_ok=True)

    LOG.debug('Acquiring lock "%(lock)s"', {'lock': name})

    try:
        ext_lock = lockutils.external_lock(name,
                                           lock_file_prefix='tobiko',
                                           lock_path=lock_path)
        gotten = ext_lock.acquire(delay=0.01, blocking=True)
        if not gotten:
            raise lockutils.AcquireLockFailedException(name)
        LOG.debug('Acquired external semaphore "%(lock)s"',
                  {'lock': name})
        try:
            yield ext_lock
        finally:
            ext_lock.release()
    finally:
        LOG.debug('Releasing lock "%(lock)s"', {'lock': name})



# Copyright (c) 2025 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Test OVN Load Balancer Database Synchronization

This test verifies that the octavia-ovn-db-sync-util can properly restore
OVN load balancers after they are deleted from the OVN Northbound database.
"""
from __future__ import absolute_import

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko import podified
from tobiko import tripleo
from tobiko.shell import sh
from tobiko.tripleo import containers


LOG = log.getLogger(__name__)


@neutron.skip_unless_is_ovn()
@keystone.skip_if_missing_service(name='octavia')
@tripleo.skip_if_overcloud
@podified.skip_if_podified  # TODO: Remove this skip when ready for Podified
class OVNLoadBalancerSyncTest(testtools.TestCase):
    """OVN Load Balancer Database Synchronization Test

    This test creates an OVN provider load balancer, deletes it from the
    OVN Northbound database, and then runs octavia-ovn-db-sync-util to
    restore it. Finally, it verifies that the load balancer functionality
    is working correctly.

    NOTE: This test is currently designed for DevStack only.
    It is skipped on TripleO and Podified/Kubernetes deployments.

    TODO: Enable for Podified environments once fully tested.
    """
    lb = None
    listener = None
    pool = None
    server_stack = tobiko.required_fixture(
        stacks.OctaviaServerStackFixture)
    other_server_stack = tobiko.required_fixture(
        stacks.OctaviaOtherServerStackFixture)

    def setUp(self):
        # pylint: disable=no-member
        super(OVNLoadBalancerSyncTest, self).setUp()

        LOG.info("Setting up OVN Load Balancer test")

        # Deploy a complete OVN load balancer
        self.lb, self.listener, self.pool = octavia.deploy_ipv4_ovn_lb(
            servers_stacks=[self.server_stack, self.other_server_stack]
        )

        # List pool members
        members = list(octavia.list_members(pool_id=self.pool.id))
        LOG.info(f"  Members: {len(members)} members in pool")
        if len(members) == 0:
            LOG.error("ERROR: No members were created in the pool!")
            LOG.error("This may indicate a problem with server stack "
                      "deployment")
        for idx, member in enumerate(members):
            LOG.info(f"    Member {idx+1}: {member.name} (ID: {member.id}, "
                     f"Address: {member.address})")

        # Wait for the load balancer to be fully operational
        octavia.wait_for_status(
            object_id=self.lb.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.ONLINE,
            get_client=octavia.get_load_balancer
        )
        LOG.info(f"Load balancer {self.lb.name} is ONLINE")

        # Verify initial traffic works
        LOG.info("Verifying initial load balancer traffic")
        octavia.verify_lb_traffic(
            pool_id=self.pool.id,
            ip_address=self.lb.vip_address,
            lb_algorithm=self.pool.lb_algorithm,
            protocol=self.listener.protocol,
            port=self.listener.protocol_port,
            timeout=300.)
        LOG.info("Load balancer traffic verified successfully")

    def _get_ovn_controller_node(self):
        """Get a controller node for executing OVN commands"""
        controllers = topology.list_openstack_nodes(group='controller')
        if not controllers:
            self.skipTest("No controller nodes found")
        return controllers[0]

    def _get_ovn_nb_connection(self):
        """Get OVN Northbound database connection string

        Tries multiple methods to obtain the connection:
        1. From octavia.conf [ovn] section using topology infrastructure
        2. From ovs-vsctl as fallback (for environments without config file)
        """
        controller = self._get_ovn_controller_node()

        # Method 1: Try to get from octavia.conf using topology
        try:
            nb_connection = topology.get_config_setting(
                file_name='octavia.conf',
                ssh_client=controller.ssh_client,
                param='ovn_nb_connection',
                section='ovn')
            if nb_connection:
                LOG.info("Found OVN NB connection from octavia.conf")
                return nb_connection
        except Exception as e:
            LOG.warning(f"Could not read from octavia.conf: {e}")

        # Method 2: Try ovs-vsctl as fallback
        try:
            LOG.info("Trying ovs-vsctl to get OVN NB connection")
            cmd = ("ovs-vsctl get open . external_ids:ovn-remote | "
                   "sed 's/\"//g' | "
                   "sed 's/6642/6641/g'")
            output = sh.execute(cmd, ssh_client=controller.ssh_client,
                                sudo=True)
            connection = output.stdout.strip()
            if connection:
                LOG.info("Found connection using ovs-vsctl")
                return connection
        except sh.ShellCommandFailed:
            pass

        self.fail("Could not determine OVN NB connection string")

    def _wrap_ovn_nbctl_command(self, ovn_nbctl_cmd):
        current_topology = topology.get_openstack_topology()
        if current_topology.has_containers:
            runtime_name = containers.get_container_runtime_name()
            cmd = f'{runtime_name} exec ovn_controller {ovn_nbctl_cmd}'
            LOG.debug("Using containerized ovn-nbctl command")
        else:
            # Direct command for non-containerized environments (DevStack)
            cmd = ovn_nbctl_cmd
            LOG.debug("Using direct ovn-nbctl command")
        return cmd

    def _get_ovn_lb_uuid(self):
        """Get the OVN Load Balancer UUID from OVN NB database"""
        controller = self._get_ovn_controller_node()
        nb_connection = self._get_ovn_nb_connection()

        # Build the ovn-nbctl command
        ovn_nbctl_cmd = (f'ovn-nbctl --db={nb_connection} --bare '
                         f'--columns=_uuid find load_balancer '
                         f'name={self.lb.id}')

        cmd = self._wrap_ovn_nbctl_command(ovn_nbctl_cmd)
        LOG.info(f"Getting OVN LB UUID with command: {cmd}")
        output = sh.execute(cmd, ssh_client=controller.ssh_client, sudo=True)
        ovn_lb_uuid = output.stdout.strip()

        if not ovn_lb_uuid:
            self.fail(f"OVN Load Balancer UUID not found for {self.lb.id}")

        LOG.info(f"Found OVN LB UUID: {ovn_lb_uuid}")
        return ovn_lb_uuid

    def _delete_ovn_lb_from_db(self, ovn_lb_uuid):
        """Delete the OVN Load Balancer from the OVN NB database"""
        controller = self._get_ovn_controller_node()
        nb_connection = self._get_ovn_nb_connection()

        # Build the ovn-nbctl command
        ovn_nbctl_cmd = f'ovn-nbctl --db={nb_connection} lb-del {ovn_lb_uuid}'

        cmd = self._wrap_ovn_nbctl_command(ovn_nbctl_cmd)
        LOG.warning(f"Deleting OVN LB from database with command: {cmd}")
        sh.execute(cmd, ssh_client=controller.ssh_client, sudo=True)
        LOG.info(f"Successfully deleted OVN LB {ovn_lb_uuid} from OVN NB DB")

    def _run_ovn_octavia_sync_tool(self):
        """Execute octavia-ovn-db-sync-util to restore the load balancer

        This method delegates to the topology class which knows how to run
        the sync tool in its environment (DevStack, Podified, TripleO, etc.)
        """
        LOG.info("Step 5: Running octavia-ovn-db-sync-tool")
        controller = self._get_ovn_controller_node()

        # Get the current topology and call its sync method
        current_topology = topology.get_openstack_topology()
        current_topology.run_octavia_ovn_db_sync(
            ssh_client=controller.ssh_client)

    def test_ovn_lb_sync_after_db_deletion(self):
        """Test OVN LB restoration after deletion from OVN NB database

        This test:
        1. Creates an OVN load balancer (done in setUp)
        2. Verifies initial functionality
        3. Gets the OVN LB UUID from OVN NB database
        4. Deletes the OVN LB from OVN NB database
        5. Runs octavia-ovn-db-sync-util to restore it
        6. Verifies the load balancer is restored and functional
        """

        # Step 1: Get OVN LB UUID
        LOG.info("Step 1: Getting OVN Load Balancer UUID")
        ovn_lb_uuid = self._get_ovn_lb_uuid()

        # Step 2: Delete OVN LB from OVN NB database
        LOG.info("Step 2: Deleting OVN LB from OVN Northbound database")
        self._delete_ovn_lb_from_db(ovn_lb_uuid)

        # Step 3: Verify LB is missing from OVN but still exists in Octavia
        LOG.info("Step 3: Verifying LB is deleted from OVN but exists in "
                 "Octavia")
        lb_in_octavia = octavia.get_load_balancer(self.lb.id)
        self.assertIsNotNone(lb_in_octavia,
                             "Load balancer should still exist in Octavia")
        LOG.info(f"Confirmed: LB {self.lb.id} still exists in Octavia DB")

        # Step 4: Run sync tool to restore the OVN LB
        LOG.info("Step 4: Running octavia-ovn-db-sync-util")
        self._run_ovn_octavia_sync_tool()

        # Step 5: Wait for LB to be restored
        LOG.info("Step 5: Waiting for load balancer to be restored")
        octavia.wait_for_status(
            object_id=self.lb.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.ONLINE,
            get_client=octavia.get_load_balancer,
            timeout=180
        )
        LOG.info(f"Load balancer {self.lb.name} is ONLINE after sync")

        # Step 6: Verify OVN LB is back in OVN NB database
        LOG.info("Step 6: Verifying OVN LB is restored in OVN NB database")
        restored_ovn_lb_uuid = self._get_ovn_lb_uuid()
        self.assertIsNotNone(restored_ovn_lb_uuid,
                             "OVN LB should be restored in OVN NB database")
        LOG.info(f"Confirmed: OVN LB restored with UUID: "
                 f"{restored_ovn_lb_uuid}")

        # Step 7: Verify traffic works again
        LOG.info("Step 7: Verifying load balancer traffic after restoration")
        octavia.verify_lb_traffic(
            pool_id=self.pool.id,
            ip_address=self.lb.vip_address,
            lb_algorithm=self.pool.lb_algorithm,
            protocol=self.listener.protocol,
            port=self.listener.protocol_port,
            timeout=300.)
        LOG.info("SUCCESS: Load balancer fully restored and functional")



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.stacks import _cirros
from tobiko.openstack.stacks import _designate
from tobiko.openstack.stacks import _l3ha
from tobiko.openstack.stacks import _manila
from tobiko.openstack.stacks import _neutron
from tobiko.openstack.stacks import _nova
from tobiko.openstack.stacks import _octavia
from tobiko.openstack.stacks import _qos
from tobiko.openstack.stacks import _advanced_vm
from tobiko.openstack.stacks import _vlan


CirrosFlavorStackFixture = _cirros.CirrosFlavorStackFixture
CirrosImageFixture = _cirros.CirrosImageFixture
CirrosServerStackFixture = _cirros.CirrosServerStackFixture
CirrosNoFipServerStackFixture = _cirros.CirrosNoFipServerStackFixture
CirrosServerWithDefaultSecurityGroupStackFixture = (
    _cirros.CirrosServerWithDefaultSecurityGroupStackFixture)
CirrosShellConnection = _cirros.CirrosShellConnection
CirrosPeerServerStackFixture = _cirros.CirrosPeerServerStackFixture
CirrosDifferentHostServerStackFixture = (
    _cirros.CirrosDifferentHostServerStackFixture)
CirrosSameHostServerStackFixture = _cirros.CirrosSameHostServerStackFixture
RebootCirrosServerOperation = _cirros.RebootCirrosServerOperation
EvacuableCirrosImageFixture = _cirros.EvacuableCirrosImageFixture
EvacuableServerStackFixture = _cirros.EvacuableServerStackFixture
ExtraDhcpOptsCirrosServerStackFixture = (
    _cirros.ExtraDhcpOptsCirrosServerStackFixture)
MultiIPCirrosServerStackFixture = (
    _cirros.MultiIPCirrosServerStackFixture)

DesignateZoneStackFixture = _designate.DesignateZoneStackFixture

L3haNetworkStackFixture = _l3ha.L3haNetworkStackFixture
L3haServerStackFixture = _l3ha.L3haServerStackFixture
L3haPeerServerStackFixture = _l3ha.L3haPeerServerStackFixture
L3haDifferentHostServerStackFixture = _l3ha.L3haDifferentHostServerStackFixture
L3haSameHostServerStackFixture = _l3ha.L3haSameHostServerStackFixture

FloatingIpStackFixture = _neutron.FloatingIpStackFixture
NetworkBaseStackFixture = _neutron.NetworkBaseStackFixture
NetworkStackFixture = _neutron.NetworkStackFixture
NetworkWithNetMtuWriteStackFixture = (
    _neutron.NetworkWithNetMtuWriteStackFixture)
RouterInterfaceStackFixture = _neutron.RouterInterfaceStackFixture
RouterStackFixture = _neutron.RouterStackFixture
SecurityGroupsFixture = _neutron.SecurityGroupsFixture
StatelessSecurityGroupFixture = _neutron.StatelessSecurityGroupFixture
get_external_network = _neutron.get_external_network
has_external_network = _neutron.has_external_network
skip_unless_has_external_network = _neutron.skip_unless_has_external_network
get_floating_network_id = _neutron.get_floating_network_id
get_floating_network = _neutron.get_floating_network
ensure_router_interface = _neutron.ensure_router_interface
has_floating_network = _neutron.has_floating_network
skip_unless_has_floating_network = _neutron.skip_unless_has_floating_network
get_router_id = _neutron.get_router_id
get_router = _neutron.get_router
has_router = _neutron.has_router
skip_unless_has_router = _neutron.skip_unless_has_router


ServerStackFixture = _nova.ServerStackFixture
KeyPairStackFixture = _nova.KeyPairStackFixture
FlavorStackFixture = _nova.FlavorStackFixture
ServerGroupStackFixture = _nova.ServerGroupStackFixture
AffinityServerGroupStackFixture = _nova.AffinityServerGroupStackFixture
AntiAffinityServerGroupStackFixture = _nova.AntiAffinityServerGroupStackFixture
CloudInitServerStackFixture = _nova.CloudInitServerStackFixture

# Octavia resources: backend servers
OctaviaServerStackFixture = _octavia.OctaviaServerStackFixture
OctaviaOtherServerStackFixture = _octavia.OctaviaOtherServerStackFixture

QosNetworkStackFixture = _qos.QosNetworkStackFixture
QosPolicyStackFixture = _qos.QosPolicyStackFixture
QosServerStackFixture = _qos.QosServerStackFixture

AdvancedFlavorStackFixture = _advanced_vm.AdvancedFlavorStackFixture
AdvancedImageFixture = _advanced_vm.AdvancedImageFixture
AdvancedServerStackFixture = _advanced_vm.AdvancedServerStackFixture
AdvancedExternalServerStackFixture = \
    _advanced_vm.AdvancedExternalServerStackFixture
AdvancedPeerServerStackFixture = _advanced_vm.AdvancedPeerServerStackFixture

VlanNetworkStackFixture = _vlan.VlanNetworkStackFixture
VlanProxyServerStackFixture = _vlan.VlanProxyServerStackFixture

# Manila resources
ManilaShareFixture = _manila.ManilaShareFixture



# Copyright 2026 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
from unittest import mock

import testtools

from tobiko.common import _lockutils


class InterworkerSynchedTest(testtools.TestCase):

    def test_lock_name_includes_class_name(self):
        """Verify that interworker_synched includes fully qualified class name
        in lock"""

        captured_lock_names = []

        # Mock the lock function to capture the lock name
        def mock_lock(name):
            captured_lock_names.append(name)
            return contextlib.nullcontext()

        # Define test fixtures
        class BaseFixture:
            @_lockutils.interworker_synched('test_lock')
            def setup_fixture(self):
                return f"Setting up {self.__class__.__name__}"

        class ChildFixture1(BaseFixture):
            pass

        class ChildFixture2(BaseFixture):
            pass

        # Patch the lock function
        with mock.patch.object(_lockutils, 'lock', side_effect=mock_lock):
            # Create instances and call setup_fixture
            base = BaseFixture()
            child1 = ChildFixture1()
            child2 = ChildFixture2()

            base.setup_fixture()
            child1.setup_fixture()
            child2.setup_fixture()

        # Verify we got 3 lock names
        self.assertEqual(len(captured_lock_names), 3)

        # Verify all lock names are unique
        self.assertEqual(len(set(captured_lock_names)), 3,
                         "All lock names should be unique")

        # Verify each lock name starts with the base name and ends with
        # the class name
        self.assertTrue(captured_lock_names[0].startswith('test_lock:'))
        self.assertTrue(captured_lock_names[0].endswith('.BaseFixture'))
        self.assertTrue(captured_lock_names[1].endswith('.ChildFixture1'))
        self.assertTrue(captured_lock_names[2].endswith('.ChildFixture2'))

        # Verify the lock name includes the module path
        for lock_name in captured_lock_names:
            self.assertIn('tobiko.tests.unit.test_lockutils', lock_name)

    def test_lock_name_without_instance(self):
        """Verify that lock works when called without instance (no args)"""

        captured_lock_names = []

        # Mock the lock function to capture the lock name
        def mock_lock(name):
            captured_lock_names.append(name)
            return contextlib.nullcontext()

        # Define a function (not a method) with the decorator
        @_lockutils.interworker_synched('function_lock')
        def standalone_function():
            return "standalone"

        # Patch the lock function
        with mock.patch.object(_lockutils, 'lock', side_effect=mock_lock):
            standalone_function()

        # When called without instance, should use base lock name
        self.assertEqual(['function_lock'], captured_lock_names)

    def test_different_fixtures_get_different_locks(self):
        """Verify different fixture classes don't block each other"""

        # This test verifies the fix for the timeout issue where
        # AdvancedServerStackFixture and AdvancedExternalServerStackFixture
        # were blocked by the same lock

        captured_lock_names = []

        def mock_lock(name):
            captured_lock_names.append(name)
            return contextlib.nullcontext()

        class ServerFixture:
            @_lockutils.interworker_synched('server_setup')
            def setup_fixture(self):
                pass

        class ExternalServerFixture(ServerFixture):
            pass

        class PeerServerFixture(ServerFixture):
            pass

        with mock.patch.object(_lockutils, 'lock', side_effect=mock_lock):
            server = ServerFixture()
            external = ExternalServerFixture()
            peer = PeerServerFixture()

            server.setup_fixture()
            external.setup_fixture()
            peer.setup_fixture()

        # All three should have different lock names
        self.assertEqual(len(captured_lock_names), 3)
        self.assertEqual(len(set(captured_lock_names)), 3,
                         "All lock names should be unique")

        # Verify each lock name contains the base name and class name
        self.assertTrue(captured_lock_names[0].startswith('server_setup:'))
        self.assertTrue(captured_lock_names[0].endswith('.ServerFixture'))
        self.assertTrue(
            captured_lock_names[1].endswith('.ExternalServerFixture'))
        self.assertTrue(captured_lock_names[2].endswith('.PeerServerFixture'))

        # Verify all contain the module path
        for lock_name in captured_lock_names:
            self.assertIn('tobiko.tests.unit.test_lockutils', lock_name)



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from datetime import datetime
import typing

import netaddr
from oslo_log import log as logging
import requests

import tobiko
from tobiko import config
from tobiko.shell import custom_script
from tobiko.shell import files
from tobiko.shell import ssh

TIMEOUT = 2  # seconds


CONF = config.CONF
LOG = logging.getLogger(__name__)

HTTP_PING_SCRIPT_NAME = "tobiko_http_ping.sh"
HTTP_PING_CURL_OUTPUT_FORMAT = "%{response_code}"
HTTP_PING_SCRIPT = """
url="http://$1";
output_file=$2;

rm $output_file

while true; do
    current_time=$(/usr/bin/date +"{time_format}");
    response_code=$(/usr/bin/curl -o /dev/null -s -I -w "{response_format}" $url);
    if [ "$response_code" -ge 200 ] && [ "$response_code" -lt 500 ]; then
            response="{result_ok}";
    else
            response="{result_failed}";
    fi;
    echo "{output_result_line}" >> $output_file;
    sleep {timeout};
done;
""".format(    # noqa: E501
    time_format=custom_script.LOG_TIME_FORMAT,
    response_format=HTTP_PING_CURL_OUTPUT_FORMAT,
    result_ok=custom_script.RESULT_OK,
    result_failed=custom_script.RESULT_FAILED,
    output_result_line=custom_script.LOG_RESULT_FORMAT,
    timeout=TIMEOUT,
)


def _ensure_http_ping_script_on_server(
        ssh_client: ssh.SSHClientType = None) -> None:
    custom_script.ensure_script_is_on_server(
        HTTP_PING_SCRIPT_NAME,
        HTTP_PING_SCRIPT,
        ssh_client=ssh_client)


def get_log_dir(
        ssh_client: ssh.SSHClientType = None) -> str:
    return custom_script.get_log_dir(
        "tobiko_http_ping_results", ssh_client)


def http_ping(
        server_ip: typing.Union[str, netaddr.IPAddress]) -> dict:
    headers = {"connection": "close"}
    result = {'time': str(datetime.now())}
    url = f"http://{server_ip}"
    try:
        response = requests.head(url, headers=headers, timeout=TIMEOUT)
        if (response.status_code >= requests.codes.ok and  # noqa; pylint: disable=no-member
                response.status_code < requests.codes.bad):  # noqa; pylint: disable=no-member
            result['response'] = custom_script.RESULT_OK
        else:
            result['response'] = custom_script.RESULT_FAILED
    except requests.exceptions.RequestException:
        result['response'] = custom_script.RESULT_FAILED
    return result


def _get_http_ping_script_command(
        server_ip: typing.Union[str, netaddr.IPAddress],
        ssh_client: ssh.SSHClientType = None):
    homedir = files.get_homedir(ssh_client)
    logfile = _get_logfile_path(server_ip, ssh_client)
    return f"bash {homedir}/{HTTP_PING_SCRIPT_NAME} {server_ip} {logfile}"


def _get_logfile_name(
        server_ip: typing.Union[str, netaddr.IPAddress]) -> str:
    return f"http_ping_{server_ip}.log"


def _get_logfile_path(
        server_ip: typing.Union[str, netaddr.IPAddress],
        ssh_client: ssh.SSHClientType = None):
    logdir_name = get_log_dir(ssh_client)
    logfile_name = _get_logfile_name(server_ip)
    return f"{logdir_name}/{logfile_name}"


def _get_http_ping_pid(
        server_ip: typing.Union[str, netaddr.IPAddress],
        ssh_client: ssh.SSHClientType = None):
    processes = custom_script.get_process_pid(
        command_line=_get_http_ping_script_command(
            server_ip, ssh_client),
        ssh_client=ssh_client)
    if not processes:
        LOG.debug(f'no http ping to server {server_ip} found.')
    return processes


def check_http_ping_results(**kwargs):
    ssh_client = kwargs.get('ssh_client')
    server_ip = kwargs.get('server_ip')
    if ssh_client:
        if not server_ip:
            tobiko.fail("Server IP is required to check http ping log file.")
        # Source log file is on the guest vm so ssh_client needs to be used
        # to get it
        src_logfile = _get_logfile_path(server_ip, ssh_client)
        dst_logfile_name = _get_logfile_name(server_ip)
        # Destination is local to where Tobiko is running so no need to pass
        # ssh_client to get_log_dir() function this time
        dest_logfile = f"{get_log_dir()}/{dst_logfile_name}"
        custom_script.copy_log_file(src_logfile, dest_logfile, ssh_client)

    logfile_name = _get_logfile_name(server_ip)
    logfiles = custom_script.get_log_files(
        glob_log_pattern=f"{get_log_dir()}/{logfile_name}")

    custom_script.check_results(logfiles)


def start_http_ping_process(
        server_ip: typing.Union[str, netaddr.IPAddress],
        ssh_client: ssh.SSHClientType = None):
    # ensure bash script is on host
    # run bash script
    _ensure_http_ping_script_on_server(ssh_client)
    if http_ping_process_alive(server_ip, ssh_client):
        return
    custom_script.start_script(
        _get_http_ping_script_command(
            server_ip, ssh_client),
        ssh_client=ssh_client)


def stop_http_ping_process(
        server_ip: typing.Union[str, netaddr.IPAddress],
        ssh_client: ssh.SSHClientType = None):
    pid = _get_http_ping_pid(server_ip, ssh_client)
    if pid:
        custom_script.stop_script(pid, ssh_client=ssh_client)


def http_ping_process_alive(
        server_ip: typing.Union[str, netaddr.IPAddress],
        ssh_client: ssh.SSHClientType = None):
    return bool(_get_http_ping_pid(server_ip, ssh_client))



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg, types


def setup_tobiko_config(conf):
    # pylint: disable=unused-argument
    from tobiko.podified import _openshift
    from tobiko.podified import _topology

    if _openshift.has_podified_cp():
        _topology.setup_podified_topology()


GROUP_NAME = "podified"
OPTIONS = [
    cfg.StrOpt('osp_project',
               default='openstack',
               help="Openshift project that includes the Openstack resources"),
    cfg.StrOpt('dataplane_node_ssh_key_secret',
               default='dataplane-ansible-ssh-private-key-secret',
               help="Name of the Openshift secret where ssh keys to connect "
                    "to EDPM nodes can be obtained from."),
    cfg.StrOpt('background_tasks_project',
               default='tobiko',
               help='Name of the OpenShift project which will be used to run '
                    'PODs with tobiko background commands, like e.g.'
                    '`tobiko ping`'),
    cfg.StrOpt('tobiko_image',
               default='quay.io/podified-antelope-centos9/openstack-tobiko:current-podified',  # noqa
               help='Contaniner image used to run background tobiko commands '
                    'like e.g. `tobiko ping` in the POD.'),
    cfg.StrOpt('iperf3_image',
               default='quay.io/podified-antelope-centos9/openstack-tobiko:current-podified',  # noqa
               help='Container image to run iperf3 client or server in the '
                    'backgroun in POD. It can be any image which provides '
                    'iperf3 but it should be in version 3.17 at least as '
                    'this version supports "--json-stream" option required '
                    'by Tobiko.'),
    cfg.IntOpt('tobiko_start_pod_timeout',
               default=300,
               help='Defines how long Tobiko will wait until POD with the '
                    'background command (like tobiko ping) will be `Running`. '
                    'In most cases, if tobiko image is already in the local '
                    'registry it will need just few seconds to start POD but '
                    'if image is not yet cached locally it may take a bit '
                    'longer time to download it.'),
    cfg.ListOpt('tobiko_pod_tolerations',
                default=[],
                bounds=True,
                item_type=types.Dict(bounds=True),
                help='List of tolerations that have to be applied to the '
                     'tobiko background pod. It is hence a list of '
                     'dictionaries. No nested disctionaries can be used. '
                     'The list has to be bound by [] and each dict has to '
                     'be bound by {}. Example: [{effect: NoSchedule, key: testOperator, value: true}, {effect: NoExecute, key: testOperator, value: true}]'),  # noqa
    cfg.StrOpt('tobiko_pod_extra_network',
               default=None,
               help='Extra network interface that needs to be attached to the '
                    'tobiko background pod.'),
    cfg.DictOpt('tobiko_pod_node_selector',
                default={},
                help='Configuration that has to be added to the tobiko '
                     'background pod in order to select a specific OCP node.'
                     ' The provided value has to be a non-nested dictionary '
                     'without any {} bouds. '
                     'Example: kubernetes.io/hostname:worker-3'),
    cfg.StrOpt('edpm_iscsid_service_name',
               default='iscsid',
               help='Specify the iscsid service name, which differs depending '
                    'on the version of the Podified Openstack environment.'),
    cfg.ListOpt('compute_dp_service_names',
                default=['nova', 'nova-custom', 'nova-custom-ceph'],
                help='List of compute dataplane service names used to '
                     'identify compute nodes in the dataplane nodesets.')
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
from __future__ import absolute_import

from oslo_log import log
import testtools

from tobiko.tests.faults.ha import test_cloud_recovery
from tobiko.tests.faults.podified.ha import cloud_disruptions
from tobiko.openstack import tests
from tobiko import podified
from tobiko.openstack import nova


LOG = log.getLogger(__name__)


def podified_health_checks():
    nova.check_nova_services_health()
    tests.test_alive_agents_are_consistent_along_time()
    # create a unique stack that will be cleaned up at the end of each test
    # TODO(eolivare) add tests.test_server_creation_no_fip() when BGP is
    # configured with expose_tenant_networks
    tests.test_server_creation()
    nova.action_on_all_instances('active')
    nova.check_virsh_domains_running()
    test_cloud_recovery.octavia_health_checks()


class PodifiedCloudHealthCheck(test_cloud_recovery.OvercloudHealthCheck):
    def setup_fixture(self):
        # run validations
        LOG.info("Start executing Podified health checks.")
        podified_health_checks()
        LOG.info("Podified health checks successfully executed.")


@podified.skip_if_not_podified
class DisruptPodifiedNodesTest(testtools.TestCase):
    """ HA Tests: run health check -> disruptive action -> health check
    disruptive_action: a function that runs some
    disruptive scenario on a node"""

    def test_0vercloud_health_check(self):
        PodifiedCloudHealthCheck.run_before()

    def test_kill_all_galera_services(self):
        PodifiedCloudHealthCheck.run_before()
        cloud_disruptions.kill_all_galera_services()
        PodifiedCloudHealthCheck.run_after()

    def test_remove_all_grastate_galera(self):
        PodifiedCloudHealthCheck.run_before()
        cloud_disruptions.remove_all_grastate_galera()
        PodifiedCloudHealthCheck.run_after()

    def test_remove_one_grastate_galera(self):
        PodifiedCloudHealthCheck.run_before()
        cloud_disruptions.remove_one_grastate_galera()
        PodifiedCloudHealthCheck.run_after()



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import re
import random

import openshift_client as oc
from oslo_log import log

import tobiko
from tobiko import config
from tobiko import podified
from tobiko.openstack import keystone

CONF = config.CONF
LOG = log.getLogger(__name__)

kill_galera = 'kill -9 $(pidof mysqld)'
rm_grastate = 'rm -rf /var/lib/mysql/grastate.dat'
galera_cluster_size = 'mysql -u root --password={passwd} -e \'SHOW STATUS ' \
                      'LIKE "wsrep_cluster_size"\''
check_bootstrap = """
ps -eo lstart,cmd | grep -v grep|
grep wsrep-cluster-address=gcomm://
"""


class GaleraBoostrapException(tobiko.TobikoException):
    message = "Bootstrap has not been activated"


class DownException(tobiko.TobikoException):
    message = "The resource is not down"


class RestoredException(tobiko.TobikoException):
    message = "The resource is not restored"


@functools.lru_cache()
def get_galera_pods_per_service(galera_service):
    # the aim of this function is just to cache results and avoid sending
    # oc requests every time
    return podified.get_pods({'service': galera_service})


def kill_all_galera_services():
    """kill all galera processes,
    check in pacemaker it is down"""
    # get galera pods sorted into 2 different lists:
    # one with 'cell-galera' an one without
    for galera_service in ('openstack-cell1-galera', 'openstack-galera'):
        pods = get_galera_pods_per_service(galera_service)
        kill_all_galera_pods(pods)
        check_all_galera_cells_down(pods[0].name())
        verify_all_galera_cells_restored(pods)


def remove_all_grastate_galera():
    """shut down galera properly,
    remove all grastate"""
    for galera_service in ('openstack-cell1-galera', 'openstack-galera'):
        pods = get_galera_pods_per_service(galera_service)
        for pod in pods:
            remove_grastate(pod.name())
        # TODO: change kill to graceful stop/ scale down
        kill_all_galera_pods(pods)
        check_all_galera_cells_down(pods[0].name())
        verify_all_galera_cells_restored(pods)


def remove_one_grastate_galera():
    """shut down galera properly,
    delete /var/lib/mysql/grastate.dat in a random node,
    check that bootstrap is done from a node with grastate"""
    for galera_service in ('openstack-cell1-galera', 'openstack-galera'):
        pods = get_galera_pods_per_service(galera_service)
        random_pod_name = random.choice(pods).name()
        remove_grastate(random_pod_name)
        # TODO: change kill to graceful stop/ scale down
        kill_all_galera_pods(pods)
        check_all_galera_cells_down(pods[0].name())
        verify_all_galera_cells_restored(pods)
        # gcomm:// without args means that bootstrap is done from this node
        bootstrap = podified.execute_in_pod(
            random_pod_name, check_bootstrap, 'galera').out().strip()
        if len(pods) > 1:
            if re.search(r'wsrep-cluster-address=gcomm://(?:\s|$)', bootstrap
                         ) is None:
                raise GaleraBoostrapException()
        elif re.search(r'wsrep-cluster-address=gcomm://', bootstrap) is None:
            raise GaleraBoostrapException()
        lastDate = re.findall(r"\w{,3}\s*\w{,3}\s*\d{,2}\s*\d{,2}:\d{,2}"
                              r":\d{,2}\s*\d{4}", bootstrap)[-1]
        LOG.info(f'last boostrap required at {lastDate}')


def remove_grastate(pod_name):
    podified.execute_in_pod(pod_name, rm_grastate, 'galera')
    LOG.info(f'grastate.dat removed from {pod_name}')


def kill_all_galera_pods(galera_pods):
    for pod in galera_pods:
        podified.execute_in_pod(pod.name(), kill_galera, 'galera')
        LOG.info(f'kill galera pod {pod}')


def check_all_galera_cells_down(pod_name):
    pw = keystone.keystone_credentials().password

    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        try:
            cluster_size = podified.execute_in_pod(
                pod_name, galera_cluster_size.format(passwd=pw), 'galera')
            error_output = cluster_size.err()
            non_error_message = "Defaulted container \"galera\" out of:"\
                                "galera, mysql-bootstrap (init)\n"
            filtered_err_output = error_output.replace(non_error_message, '')
            if not filtered_err_output.strip():
                continue
        except oc.OpenShiftPythonException:
            LOG.info('all galera cells down')
            return
    raise DownException()


def verify_all_galera_cells_restored(pods):
    pw = keystone.keystone_credentials().password

    retry = tobiko.retry(timeout=160, interval=10)
    for _ in retry:
        pod_name = pods[0].name()
        try:
            cluster_size = podified.execute_in_pod(
                pod_name, galera_cluster_size.format(passwd=pw), 'galera')
        except oc.OpenShiftPythonException:
            continue

        wsrep_cluster_size = int(re.search(r'wsrep_cluster_size\s+(\d+)',
                                           cluster_size.out()).group(1))
        if wsrep_cluster_size == len(pods):
            LOG.info('all galera cells are restored')
            return

    raise RestoredException()



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.rhosp import _version_utils
from tobiko.rhosp import _topology

RhospTopology = _topology.RhospTopology
RhospNode = _topology.RhospNode

get_rhosp_release = _version_utils.get_rhosp_release
get_rhosp_version = _version_utils.get_rhosp_version

ip_to_hostname = _topology.ip_to_hostname



# Copyright (c) 2025 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# mypy: disable-error-code="attr-defined"
from __future__ import absolute_import

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import octavia
from tobiko.openstack import nova
from tobiko.openstack import stacks


LOG = log.getLogger(__name__)


@neutron.skip_unless_is_ovn()
@keystone.skip_if_missing_service(name='octavia')
class OctaviaOVNProviderHealthMonitorTest(testtools.TestCase):
    # pylint: disable=no-member
    """Octavia OVN provider health monitor test.

    Create an OVN provider load balancer with 2 members.
    Create a client that is connected to the load balancer
    Create a health monitor that is connected to the load balancer,
    pause server one validate degraded status
    pause other server validate error status
    bring both servers back and validate status
    """
    lb = None
    listener = None
    pool = None
    health_monitor = None
    server_stack = tobiko.required_fixture(
        stacks.OctaviaServerStackFixture)
    other_server_stack = tobiko.required_fixture(
        stacks.OctaviaOtherServerStackFixture)

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaOVNProviderHealthMonitorTest, self).setUp()

        self.lb, self.listener, self.pool = octavia.deploy_ipv4_ovn_lb(
            servers_stacks=[self.server_stack, self.other_server_stack]
        )

        self.health_monitor = octavia.deploy_hm(
            octavia.OVN_PROVIDER, octavia.HM_OVN_NAME, self.pool.id)

    def test_hm(self) -> None:
        # Wait for health monitor to be ONLINE
        octavia.wait_for_status(
            object_id=self.health_monitor.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.ONLINE,
            get_client=octavia.get_health_monitor
        )
        LOG.info(f"Health monitor {self.health_monitor.name} is ONLINE")

        # Wait for load balancer to be ONLINE
        octavia.wait_for_status(
            object_id=self.lb.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.ONLINE,
            get_client=octavia.get_load_balancer
        )
        LOG.info(f"Load balancer {self.lb.name} is ONLINE")

        # Stop first server and wait for DEGRADED status
        server_one = nova.find_server(
            id=self.server_stack.outputs.server_id)
        other_server = nova.find_server(
            id=self.other_server_stack.outputs.server_id)

        server_one.stop()
        nova.wait_for_server_status(server=server_one.id, timeout=900,
                                    status='SHUTOFF')
        octavia.wait_for_status(
            object_id=self.lb.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.DEGRADED,
            get_client=octavia.get_load_balancer
        )
        LOG.info(f"Load balancer {self.lb.name} is DEGRADED after pausing "
                 "first server")

        # Stop second server and wait for ERROR status
        other_server.stop()
        nova.wait_for_server_status(server=other_server.id,  timeout=900,
                                    status='SHUTOFF')
        octavia.wait_for_status(
            object_id=self.lb.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.ERROR,
            get_client=octavia.get_load_balancer
        )
        LOG.info(f"Load balancer {self.lb.name} is ERROR after pausing both "
                 "servers")

        # Start second server and wait for DEGRADED status
        other_server.start()
        nova.wait_for_server_status(server=other_server.id,  timeout=900,
                                    status='ACTIVE')

        octavia.wait_for_status(
            object_id=self.lb.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.DEGRADED,
            get_client=octavia.get_load_balancer
        )
        LOG.info(f"Load balancer {self.lb.name} is DEGRADED after unpausing "
                 "second server")

        # Start first server and wait for ONLINE status
        server_one.start()
        nova.wait_for_server_status(server=server_one.id,  timeout=900,
                                    status='ACTIVE')
        octavia.wait_for_status(
            object_id=self.lb.id,
            status_key=octavia.OPERATING_STATUS,
            status=octavia.ONLINE,
            get_client=octavia.get_load_balancer
        )
        LOG.info(f"Load balancer {self.lb.name} is ONLINE after unpausing "
                 "both servers")



from __future__ import absolute_import

import abc
import os
import re
import typing

from oslo_log import log

import tobiko
from tobiko.openstack import topology
from tobiko import podman
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


expected_containers_file = os.path.expanduser(
    '~/expected_containers_list_td.csv')


class ContainerRuntime(abc.ABC):
    runtime_name: str
    version_pattern: typing.Pattern

    def match_version(self, version: str) -> bool:
        for version_line in version.splitlines():
            if self.version_pattern.match(version_line) is not None:
                return True
        return False

    def get_client(self, ssh_client):
        for attempt in tobiko.retry(timeout=60.0,
                                    interval=5.0):
            try:
                client = self._get_client(ssh_client=ssh_client)
                break
            # TODO chose a better exception type
            except Exception:
                if attempt.is_last:
                    raise
                LOG.debug(f'Unable to connect to {self.runtime_name} server',
                          exc_info=1)
                ssh.reset_default_ssh_port_forward_manager()
        else:
            raise RuntimeError("Broken retry loop")
        return client

    def _get_client(self, ssh_client):
        raise NotImplementedError

    def list_containers(self, ssh_client):
        raise NotImplementedError


# NOTE: only podman is supported, but we still maintian the base abstract class
# ContainerRuntime in case a different container runtime is supported in the
# future
class PodmanContainerRuntime(ContainerRuntime):
    runtime_name = 'podman'
    version_pattern = re.compile('Podman version .*', re.IGNORECASE)

    def _get_client(self, ssh_client):
        return podman.get_podman_client(ssh_client=ssh_client).connect()

    def list_containers(self, ssh_client):
        client = self.get_client(ssh_client=ssh_client)
        return podman.list_podman_containers(client=client)


PODMAN_RUNTIME = PodmanContainerRuntime()
CONTAINER_RUNTIMES = [PODMAN_RUNTIME]


class ContainerMismatchException(tobiko.TobikoException):
    pass


def remove_containers_from_comparison(comparable_containers_td):
    """remove any containers if comparing them with previous status is not
    necessary or makes no sense
    """
    final_td = tobiko.TableData()
    os_topology = topology.get_openstack_topology()

    for item in comparable_containers_td.data:
        add_this_item = True
        for ignore_container in os_topology.ignore_containers_list:
            if ignore_container in item['container_name']:
                LOG.info(f'container {ignore_container} has changed state, '
                         'but that\'s ok - it will be ignored and the test '
                         f'will not fail due to this: {item}')
                # if a pcs resource is found , we drop that row
                add_this_item = False
                # this row was already dropped, go to next row
                break
        if add_this_item:
            final_td.append(item)
    return final_td


def tabledata_difference(td1, td2, which=None):
    """Find rows which are different between two TableData objects."""
    # Check if schemas match
    if td1.schema != td2.schema:
        raise ValueError("TableData objects must have the same schema")

    # Convert rows to hashable format for comparison
    def row_to_hashable(row):
        return tuple(sorted(row.items()))

    # Create sets of hashable rows
    td1_rows = {row_to_hashable(row): row for row in td1.data}
    td2_rows = {row_to_hashable(row): row for row in td2.data}

    # Find common rows, left-only rows, and right-only rows
    common_keys = set(td1_rows.keys()) & set(td2_rows.keys())
    left_only_keys = set(td1_rows.keys()) - set(td2_rows.keys())
    right_only_keys = set(td2_rows.keys()) - set(td1_rows.keys())

    # Create result rows with indicator column
    result_rows = []

    # Add common rows
    for key in common_keys:
        row = td1_rows[key].copy()
        row['same_state'] = 'both'
        result_rows.append(row)

    # Add left-only rows
    for key in left_only_keys:
        row = td1_rows[key].copy()
        row['same_state'] = 'left_only'
        result_rows.append(row)

    # Add right-only rows
    for key in right_only_keys:
        row = td2_rows[key].copy()
        row['same_state'] = 'right_only'
        result_rows.append(row)

    # Create result TableData
    comparison_td = tobiko.TableData(result_rows)

    # Filter based on which parameter (similar to original function)
    if which is None:
        # Return only non-identical rows
        diff_td = tobiko.TableData([row for row in comparison_td.data
                                    if row['same_state'] != 'both'])
    else:
        # Return only rows with specified state
        diff_td = tobiko.TableData([row for row in comparison_td.data
                                    if row['same_state'] == which])

    # Apply the same filtering logic as the original function
    final_diff_td = remove_containers_from_comparison(diff_td)

    return final_diff_td



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import glob
import json
import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.shell import http_ping
from tobiko.shell import iperf3
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.shell import ssh

CONF = config.CONF
LOG = log.getLogger(__name__)

OSP_CONTROLPLANE = 'openstackcontrolplane'
OSP_DP_NODESET = 'openstackdataplanenodeset'
OSP_CONFIG_SECRET_NAME = 'secret/openstack-config-secret'
OSP_BM_HOST = 'baremetalhost.metal3.io'
OSP_BM_CRD = 'baremetalhosts.metal3.io'
OCP_NODES = 'nodes'
OVNDBCLUSTER = 'ovndbcluster'

OVN_DP_SERVICE_NAME = 'ovn'

EDPM_COMPUTE_GROUP = 'edpm-compute'
EDPM_NETWORKER_GROUP = 'edpm-networker'
EDPM_OTHER_GROUP = 'edpm-other'


_IS_OC_CLIENT_AVAILABLE = None
_IS_BM_CRD_AVAILABLE = None
_TOBIKO_PROJECT_EXISTS = None

try:
    import openshift_client as oc
except ModuleNotFoundError:
    _IS_OC_CLIENT_AVAILABLE = False

# NOTE(slaweq): This path is "hardcoded" in the tobiko.shell.ping._ping
# module currently so lets use it here like that as well. Maybe in the
# future there will be need to make it configurable but for now it is
# not needed.
PING_RESULTS_DIR = 'tobiko_ping_results'
# Also directory where results are stored inside the POD is hardcoded,
# It is in the $HOME/{PING_RESULTS_DIR}/ and $HOME inside the tobiko container
# is "/var/lib/tobiko"
POD_PING_RESULTS_DIR = f"/var/lib/tobiko/{PING_RESULTS_DIR}"

HTTP_PING_RESULTS_DIR = "tobiko_http_ping_results"
POD_HTTP_PING_RESULTS_DIR = f"/var/lib/tobiko/{HTTP_PING_RESULTS_DIR}"


def _is_oc_client_available() -> bool:
    # pylint: disable=global-statement
    global _IS_OC_CLIENT_AVAILABLE
    if _IS_OC_CLIENT_AVAILABLE is None:
        _IS_OC_CLIENT_AVAILABLE = False
        try:
            if sh.execute('which oc').exit_status == 0:
                _IS_OC_CLIENT_AVAILABLE = True
        except sh.ShellCommandFailed:
            pass
    return _IS_OC_CLIENT_AVAILABLE


def _is_baremetal_crd_available() -> bool:
    # pylint: disable=global-statement
    global _IS_BM_CRD_AVAILABLE
    if not _is_oc_client_available():
        return False
    if _IS_BM_CRD_AVAILABLE is None:
        try:
            # oc.selector("crd") does not need to run on a specific OCP project
            _IS_BM_CRD_AVAILABLE = any(
                [OSP_BM_CRD in n for n in oc.selector("crd").qnames()])
        except oc.OpenShiftPythonException:
            _IS_BM_CRD_AVAILABLE = False
    return _IS_BM_CRD_AVAILABLE


def _get_group(services):
    for compute_dp_service in CONF.tobiko.podified.compute_dp_service_names:
        if compute_dp_service in services:
            return EDPM_COMPUTE_GROUP
    if OVN_DP_SERVICE_NAME in services:
        return EDPM_NETWORKER_GROUP
    return EDPM_OTHER_GROUP


def _get_ocp_node_hostname(node):
    for address in node.get('status', {}).get('addresses', []):
        if address.get('type') == 'Hostname':
            return address['address']


def _get_ocp_node_addresses(node):
    return [
        netaddr.IPAddress(address['address']) for
        address in node.get('status', {}).get('addresses', [])
        if address.get('type') != 'Hostname']


def _get_ocp_node_roles(node_dict):
    # Gets a node_dict; returns a list of node's roles
    labels = node_dict.get('metadata', {}).get('labels', {})
    return [
        key for key in labels.keys()
        if key.startswith("node-role.kubernetes.io/")
    ]


def _get_ocp_node_taints(node_dict):
    # Gets a node_dict; returns a list of node's taints
    return node_dict.get("spec", {}).get("taints", []) or []


def _get_edpm_node_ctlplane_ip_from_status(hostname, node_status):
    all_ips = node_status.get('AllIPs') or node_status.get('allIPs')
    if not all_ips:
        LOG.warning("No IPs found in the Nodeset status: %s",
                    node_status)
        return
    host_ips = all_ips.get(hostname)
    if not host_ips:
        LOG.warning("Host %s not found in AllIPs: %s",
                    hostname, all_ips)
        return
    return host_ips.get('ctlplane')


def has_podified_cp() -> bool:
    if not _is_oc_client_available():
        LOG.debug("Openshift CLI client isn't installed.")
        return False
    try:
        return bool(
            oc.selector(OSP_CONTROLPLANE, all_namespaces=True).objects())
    except oc.OpenShiftPythonException:
        return False


def get_dataplane_ssh_keypair():
    private_key = ""
    public_key = ""
    secret_name = (
        f'secret/{CONF.tobiko.podified.dataplane_node_ssh_key_secret}')
    try:
        with oc.project(CONF.tobiko.podified.osp_project):
            secret_object = oc.selector(secret_name).object()
        private_key = secret_object.as_dict(
            ).get('data', {}).get('ssh-privatekey', "")
        public_key = secret_object.as_dict(
            ).get('data', {}).get('ssh-publickey', "")
    except oc.OpenShiftPythonException as err:
        LOG.error("Error while trying to get Dataplane secret SSH Key: %s",
                  err)
    return private_key, public_key


def get_openstack_config_secret():
    with oc.project(CONF.tobiko.podified.osp_project):
        try:
            secret_object = oc.selector(OSP_CONFIG_SECRET_NAME).object()
        except oc.OpenShiftPythonException as err:
            LOG.info("Error while trying to get openstack config secret "
                     f"{OSP_CONFIG_SECRET_NAME} from Openshift. Error: {err}")
            return
    return secret_object.as_dict()


def list_edpm_nodes():
    nodes = []
    with oc.project(CONF.tobiko.podified.osp_project):
        nodesets = oc.selector(OSP_DP_NODESET).objects()
    for nodeset in nodesets:
        nodeset_spec = nodeset.as_dict()['spec']
        nodeset_status = nodeset.as_dict()['status']
        node_template = nodeset_spec['nodeTemplate']
        nodeset_nodes = nodeset_spec['nodes']
        group_name = _get_group(nodeset_spec['services'])
        for node in nodeset_nodes.values():
            node_hostname = node.get('hostName')
            node_dict = {
                'hostname': node_hostname,
                'host': (node['ansible'].get('ansibleHost') or
                         _get_edpm_node_ctlplane_ip_from_status(
                             node_hostname, nodeset_status)),
                'group': group_name,
                'port': (
                    node.get('ansible', {}).get('ansiblePort') or
                    node_template.get('ansible', {}).get('ansiblePort')),
                'username': (
                    node.get('ansible', {}).get('ansibleUser') or
                    node_template.get('ansible', {}).get('ansibleUser')),
            }
            nodes.append(node_dict)
    return nodes


def list_ocp_nodes():
    # oc.selector("nodes") does not need to run on a specific OCP project
    nodes_sel = oc.selector(OCP_NODES)
    ocp_nodes = []
    for node in nodes_sel.objects():
        node_dict = node.as_dict()
        ocp_nodes.append({
            'hostname': _get_ocp_node_hostname(node_dict),
            'addresses': _get_ocp_node_addresses(node_dict),
            'roles': _get_ocp_node_roles(node_dict),
            'taints': _get_ocp_node_taints(node_dict)
        })
    return ocp_nodes


def power_on_edpm_node(nodename):
    _set_edpm_node_online_status(nodename, online=True)


def power_off_edpm_node(nodename):
    _set_edpm_node_online_status(nodename, online=False)


def _set_edpm_node_online_status(nodename, online):
    if _is_baremetal_crd_available() is False:
        LOG.info("BareMetal operator is not available in the deployment. "
                 "Starting and stopping EDPM nodes is not supported.")
        return
    try:
        with oc.project(CONF.tobiko.podified.osp_project):
            bm_node = oc.selector(f"{OSP_BM_HOST}/{nodename}").objects()[0]
    except oc.OpenShiftPythonException as err:
        LOG.info(f"Error while trying to get BareMetal Node '{nodename}' "
                 f"from Openshift. Error: {err}")
        return
    except IndexError:
        LOG.error(f"Node {nodename} not found in the {OSP_BM_HOST} CRs.")
        return
    bm_node.model.spec['online'] = online
    try:
        # NOTE(slaweq): returned status is 0 when all operations where
        # finished successfully. Otherwise status will be different than
        # 0, like in the shell scripts
        if not bool(bm_node.apply().status()):
            _wait_for_poweredOn_status(nodename, online)
    except oc.OpenShiftPythonException as err:
        LOG.error(f"Error while applying new online state: {online} for "
                  f"the node: {nodename}. Error: {err}")


def _wait_for_poweredOn_status(nodename, expected_status,
                               timeout: tobiko.Seconds = None):
    for attempt in tobiko.retry(
            timeout=timeout,
            count=10,
            interval=5.,
            default_timeout=30):
        LOG.debug(f"Checking power status of the '{nodename}'.")
        try:
            with oc.project(CONF.tobiko.podified.osp_project):
                poweredOn = oc.selector(
                    f"{OSP_BM_HOST}/{nodename}"
                ).objects()[0].model.status['poweredOn']
        except oc.OpenShiftPythonException as err:
            LOG.error("Error while trying to get 'poweredOn' state of "
                      f"the node {nodename}. Error: {err}")
        else:
            if poweredOn == expected_status:
                LOG.debug(f"Actual poweredOn state of the node {nodename} "
                          f"is: '{poweredOn}' which is as expected.")
                return True
            LOG.debug(f"Actual poweredOn state is: '{poweredOn}' != "
                      f" '{expected_status}'")
        attempt.check_limits()


def get_ovndbcluter(ovndbcluster_name):
    with oc.project(CONF.tobiko.podified.osp_project):
        ovndbcluter = oc.selector(
            f"{OVNDBCLUSTER}/{ovndbcluster_name}").objects()
    if len(ovndbcluter) != 1:
        tobiko.fail(f"Unexpected number of {OVNDBCLUSTER}/{ovndbcluster_name} "
                    f"objects obtained: {len(ovndbcluter)}")
    return ovndbcluter[0].as_dict()


def get_pods(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).objects()


def get_pod_names(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).qnames()


def get_pod_count(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).count_existing()


def delete_pods(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).delete()


def _project_exists(name):
    projects_selector = oc.selector(f"projects/{name}")
    return not len(projects_selector.objects()) == 0


def _ensure_project_exists(name):
    # pylint: disable=global-statement
    global _TOBIKO_PROJECT_EXISTS
    if not _TOBIKO_PROJECT_EXISTS and not _project_exists(name):
        project_def = {
            'apiVersion': 'v1',
            'kind': 'Namespace',
            'metadata': {
                'name': name
            }
        }
        oc.create(project_def)
        _TOBIKO_PROJECT_EXISTS = True


def tobiko_project_context():
    _ensure_project_exists(CONF.tobiko.podified.background_tasks_project)
    return oc.project(CONF.tobiko.podified.background_tasks_project)


def check_or_start_tobiko_ping_command(server_ip):
    cmd_args = ['ping', server_ip, '--interval', CONF.tobiko.ping.interval]
    pod_name = f'tobiko-ping-{server_ip}'.replace('.', '-')
    return check_or_start_tobiko_command(
        cmd_args, pod_name, _check_ping_results)


def check_or_start_tobiko_command(cmd_args, pod_name, check_function):
    pod_obj = _get_pod(pod_name)
    if pod_obj:
        # in any case test is still running, check for failures:
        # execute process check i.e. go over results file
        # truncate the log file and restart the POD with background
        # command
        LOG.info('running a check function: '
                 f'on results of processes: {pod_name}')
        check_function(pod_obj)
        with tobiko_project_context():
            pod_obj.delete(ignore_not_found=True)
            LOG.info('checked and stopped previous tobiko command '
                     f'POD {pod_name}; starting a new POD.')
    elif config.is_prevent_create():
        tobiko.fail(f'Expected POD {pod_name} not running. '
                    f'That POD should have been running: {cmd_args}')
    else:
        # First time the test is run:
        # if POD by specific name is not present start one:
        LOG.info('No previous tobiko command POD found: '
                 f'{pod_name}, starting a new POD '
                 f'of function: {cmd_args}')

    pod_obj = _start_tobiko_command_pod(cmd_args, pod_name)
    # check test is not failing from the beginning
    check_function(pod_obj)


def _get_pod(pod_name):
    with tobiko_project_context():
        pod_sel = oc.selector(f'pod/{pod_name}')
        if len(pod_sel.objects()) > 1:
            raise tobiko.MultipleObjectsFound(pod_sel.objects())
        if not pod_sel.objects():
            return
        return pod_sel.objects()[0]


def _start_pod(cmd, args, pod_name, pod_image):
    pod_def = {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {
            "name": pod_name,
            "namespace": CONF.tobiko.podified.background_tasks_project
        },
        "spec": {
            "containers": [{
                "name": pod_name,
                "image": pod_image,
                "command": cmd,
                # All items from args have to be converted to
                # strings, otherwise the pod creation may fail
                "args": [str(arg) for arg in args],
            }],
            "restartPolicy": "Never"
        }
    }

    if CONF.tobiko.podified.tobiko_pod_extra_network:
        pod_def["metadata"]["annotations"] = {
            "k8s.v1.cni.cncf.io/networks":
            CONF.tobiko.podified.tobiko_pod_extra_network}
    if CONF.tobiko.podified.tobiko_pod_tolerations:
        pod_def["spec"]["tolerations"] = \
            CONF.tobiko.podified.tobiko_pod_tolerations
    if CONF.tobiko.podified.tobiko_pod_node_selector:
        pod_def["spec"]["nodeSelector"] = \
            CONF.tobiko.podified.tobiko_pod_node_selector

    with tobiko_project_context():
        pod_sel = oc.create(pod_def)
        with oc.timeout(CONF.tobiko.podified.tobiko_start_pod_timeout):
            try:
                success, pod_objs, _ = pod_sel.until_all(
                    success_func=lambda pod:
                        pod.as_dict()['status']['phase'] == 'Running'
                )
            except oc.OpenShiftPythonException:
                LOG.error(f'Pod {pod_name} creation failed')
                pod_describe = oc.selector(
                    f'pod/{pod_name}').object().describe()
                LOG.error(f'{pod_describe}')
                raise

    if success:
        return pod_objs[0]


def _start_tobiko_command_pod(cmd_args, pod_name):
    return _start_pod(
        cmd=["tobiko"], args=cmd_args, pod_name=pod_name,
        pod_image=CONF.tobiko.podified.tobiko_image)


def _copy_file_from_pod(pod, src, dest, dest_file_pattern):
    for attempt in tobiko.retry(timeout=30., interval=5.):
        cp = oc.oc_action(pod.context, 'cp', [src, dest])
        if cp.status == 0 and glob.glob(dest_file_pattern):
            break
        elif attempt.is_last:
            tobiko.fail("Failed to copy files from the POD "
                        f"{pod.name()}. Error: {cp.err}")


def _check_ping_results(pod):
    # NOTE(slaweq): we have to put ping log files in the directory
    #   as defined below because it is expected to be like that by the
    #   tobiko.shell.ping._ping module so we can use those existing
    #   functions to check results
    ping_results_dest = f'{sh.get_user_home_dir()}/{PING_RESULTS_DIR}'
    ping_log_file_pattern = f'{ping_results_dest}/ping_*.log'
    _copy_file_from_pod(
        pod, f"{pod.name()}:{POD_PING_RESULTS_DIR}", ping_results_dest,
        ping_log_file_pattern)
    # ping.check_ping_statistics() calls tobiko.truncate_logfile(filename) to
    # rename log files to ping_<IP>.log_<date>
    ping.check_ping_statistics()


def execute_in_pod(pod_name, command, container_name=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector(f'pod/{pod_name}').object().execute(
            ['sh', '-c', command], container_name=container_name)


def _get_iperf_client_pod_name(
        address: typing.Union[str, netaddr.IPAddress]) -> str:
    return f'tobiko-iperf-client-{address}'.replace('.', '-')


def _store_iperf3_client_results(
        address: typing.Union[str, netaddr.IPAddress],
        output_dir: str = 'tobiko_iperf_results'):
    # openshift client returns logs in dict where keyname has format
    # <fully-qualified-name> -> <log output>
    # In this case, we don't really need to check it as requested logs
    # are only from the single POD so it can just try to get first
    # item from the dict's values()
    pod_obj = _get_pod(_get_iperf_client_pod_name(address))
    raw_pod_logs = list(pod_obj.logs().values())[0]
    if not raw_pod_logs:
        LOG.warning('No logs from the iperf3 client POD.')
        return

    # Logs are printed by ipef3 client to the stdout in json
    # format, but the format is different then what is stored
    # in the file when "--logfile" option is used in ipef3
    # So to be able to validate them in the same way, logs from
    # stdout of the Pod needs to be converted
    iperf3_results_data = iperf3.parse_json_stream_output(raw_pod_logs)

    logfile = iperf3.get_iperf3_logs_filepath(address, output_dir)
    with open(logfile, "w") as f:
        json.dump(iperf3_results_data, f)


def start_iperf3(
        address: typing.Union[str, netaddr.IPAddress],
        bitrate: int = None,
        download: bool = None,
        port: int = None,
        protocol: str = None,
        iperf3_server_ssh_client: ssh.SSHClientType = None,
        **kwargs):  # noqa; pylint: disable=W0613

    if iperf3_server_ssh_client:
        iperf3.start_iperf3_server(
            port, protocol, iperf3_server_ssh_client)

    parameters = iperf3.iperf3_client_parameters(
        address=address, bitrate=bitrate,
        download=download, port=port, protocol=protocol,
        timeout=0, json_stream=True)

    cmd_args = iperf3.get_iperf3_client_command(
            parameters).as_list()[1:]
    pod_name = _get_iperf_client_pod_name(address)
    _start_pod(
        cmd=["iperf3"], args=cmd_args, pod_name=pod_name,
        pod_image=CONF.tobiko.podified.iperf3_image)


def stop_iperf3_client(
        address: typing.Union[str, netaddr.IPAddress],
        **kwargs):  # noqa; pylint: disable=W0613
    # First logs from the POD needs to be stored in the file
    # so that it can be validated later
    _store_iperf3_client_results(address)

    pod_obj = _get_pod(_get_iperf_client_pod_name(address))
    with tobiko_project_context():
        pod_obj.delete(ignore_not_found=True)


def iperf3_pod_alive(
        address: typing.Union[str, netaddr.IPAddress],  # noqa; pylint: disable=W0613
        **kwargs) -> bool:
    pod_obj = _get_pod(_get_iperf_client_pod_name(address))
    if not pod_obj:
        return False
    return pod_obj.as_dict()['status']['phase'] == 'Running'


def _get_http_ping_pod_name(
        server_ip: typing.Union[str, netaddr.IPAddress]) -> str:
    return f'tobiko-http-ping-{server_ip}'.replace('.', '-')


def _check_http_ping_results_from_pod(pod):
    """Copy log file from the POD and then check file locally."""
    results_dest = http_ping.get_log_dir()
    log_file_pattern = f'{results_dest}/http_ping_*.log'
    _copy_file_from_pod(
        pod, f"{pod.name()}:{POD_HTTP_PING_RESULTS_DIR}", results_dest,
        log_file_pattern)
    http_ping.check_http_ping_results()


def check_or_start_tobiko_http_ping_command(
        server_ip: typing.Union[str, netaddr.IPAddress]):
    cmd_args = ['http', 'ping', server_ip]
    pod_name = _get_http_ping_pod_name(server_ip)
    return check_or_start_tobiko_command(
        cmd_args, pod_name, _check_http_ping_results_from_pod)


def get_ocp_node_uptime(node_name: str):
    # timeout is needed to avoid that the `oc debug` command gets stuck forever
    output = sh.execute(f"timeout 10 oc debug node/{node_name} -- "
                        "chroot /host cat /proc/uptime").stdout
    uptime_line = output.splitlines()[0]
    uptime_string = uptime_line.split()[0]
    return float(uptime_string)


def reboot_ocp_node(node_name: str):
    LOG.debug(f"Rebooting OCP node {node_name}...")
    sh.execute(f"oc debug node/{node_name} -- chroot /host reboot")
    LOG.debug(f"Reboot command sent to OCP node {node_name}.")



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import pytest
import testtools

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko.shell import ping
from tobiko.shell import sh


class BaseNetworkTest(testtools.TestCase):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(stacks.CirrosPeerServerStackFixture)


@pytest.mark.minimal
class NetworkTest(BaseNetworkTest):

    def test_stack_create_complete(self):
        self.stack.key_pair_stack.wait_for_create_complete()
        self.stack.network_stack.wait_for_create_complete()
        self.stack.peer_stack.wait_for_create_complete()
        self.stack.wait_for_create_complete()

    def test_ssh(self):
        """Test TCP connectivity to SSH server from VM to VM"""
        hostname = sh.ssh_hostname(ssh_client=self.stack.ssh_client)
        self.assertEqual(self.stack.server_name.lower(), hostname)

    def test_ping(self):
        """Test ICMP connectivity to from VM to VM"""
        ping.assert_reachable_hosts(
            [self.stack.ip_address],
            ssh_client=self.stack.peer_stack.ssh_client)

    # --- test l3_ha extension ------------------------------------------------

    @neutron.skip_unless_is_ovs()
    @neutron.skip_if_missing_networking_extensions('l3-ha')
    def test_l3_ha(self):
        """Test l3-ha network attribute"""
        gateway = self.stack.network_stack.gateway_details
        self.assertEqual(self.stack.network_stack.ha,
                         gateway['ha'])


@pytest.mark.background
class BackgroundProcessTest(BaseNetworkTest):
    """Test designed to run in the background,
    then collect results.
    Logic: checks if process exists, if so stop the process,
    then execute some check logic i.e. a check function.
    if the process by name isn't running,
    start a separate process i.e a background function"""

    stack = tobiko.required_fixture(stacks.AdvancedPeerServerStackFixture)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.topology = topology.get_openstack_topology()

    def test_check_background_vm_ping(self):
        """Ping from test machine/container/pod to VM with FIP,
        validating north-south connectivity with SDNAT (source-destination
        NAT)."""
        self.topology.check_or_start_background_vm_ping(
            self.stack.peer_stack.floating_ip_address)

    def test_check_background_vm_ping_snat(self):
        """Ping from a VM without FIP to an external IP,
        validating north-south connectivity with SNAT (source NAT)."""
        # make sure the VM does not have any FIP
        self.assertFalse(self.stack.has_floating_ip)

        try:
            ext_subnet = neutron.list_subnets(
                network=self.stack.network_stack.gateway_network_id,
                ip_version=4)[0]
        except IndexError:
            ext_subnet = neutron.list_subnets(
                network=self.stack.network_stack.gateway_network_id,
                ip_version=6)[0]

        self.topology.check_or_start_background_vm_ping(
            ext_subnet['gateway_ip'],
            ssh_client=self.stack.ssh_client)

    def test_check_background_vm_ping_east_west(self):
        """Pings between VMs conected to the same tenant network,
        validating east-west connectivity."""
        self.topology.check_or_start_background_vm_ping(
            self.stack.fixed_ipv4,
            ssh_client=self.stack.peer_stack.ssh_client)

    def test_east_west_tcp_traffic_background_iperf(self):
        """ Test East-West TCP traffic in the existing flow.

        This test is intended to test TCP traffic with bigger amount of
        data send between two VMs connected to the same tenant network.
        Traffic is send in the single flow using "iperf" tool.
        """

        self.topology.check_or_start_background_iperf_connection(
                self.stack.fixed_ipv4,
                port=5203,
                protocol='tcp',
                ssh_client=self.stack.peer_stack.ssh_client,
                iperf3_server_ssh_client=self.stack.ssh_client)

    def test_north_south_tcp_traffic_background_iperf(self):
        """ Test North-South TCP traffic in the existing flow.

        This test is intended to test TCP traffic with bigger amount of
        data send VM with Floating IP and external world.
        Traffic is send in the single flow using "iperf" tool.
        """

        self.topology.check_or_start_background_iperf_connection(
                self.stack.peer_stack.floating_ip_address,
                port=5204,
                protocol='tcp',
                iperf3_server_ssh_client=self.stack.peer_stack.ssh_client)

    def test_north_south_tcp_new_connections(self):
        self.topology.check_or_start_background_http_ping(
            self.stack.peer_stack.floating_ip_address)

    def test_east_west_tcp_new_connections(self):
        self.topology.check_or_start_background_http_ping(
            self.stack.fixed_ipv4,
            ssh_client=self.stack.peer_stack.ssh_client)

    def test_dhcp_service(self):
        """ Test constantly if VM can get response from the DHCP server"""
        self.topology.check_or_start_background_dhcp_ping(
            ssh_client=self.stack.peer_stack.ssh_client)

    def test_dns_service(self):
        """ Test constantly if VM can get response from the DNS server"""
        port_details = self.stack.port_details
        if not port_details:
            tobiko.skip_test(
                'DNS background test requires DNS information to be assigned '
                'to the ports in Neutron.')

        self.topology.check_or_start_background_dns_ping(
            ip_address=port_details['dns_assignment'][0]['ip_address'],
            fqdn=port_details['dns_assignment'][0]['fqdn'],
            ssh_client=self.stack.peer_stack.ssh_client)


@pytest.mark.migrate_server
class SameHostNetworkTest(NetworkTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.CirrosSameHostServerStackFixture)

    def test_same_host(self):
        sender = self.stack.peer_stack.server_details
        receiver = self.stack.server_details
        self.assertEqual({'same_host': [sender.id]},
                         self.stack.scheduler_hints)
        self.assertEqual(nova.get_server_hypervisor(sender),
                         nova.get_server_hypervisor(receiver))


@pytest.mark.migrate_server
@nova.skip_if_missing_hypervisors(count=2, state='up', status='enabled')
class DifferentHostNetworkTest(NetworkTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.CirrosDifferentHostServerStackFixture)

    def test_different_host(self):
        sender = self.stack.peer_stack.server_details
        receiver = self.stack.server_details
        self.assertEqual({'different_host': [sender.id]},
                         self.stack.scheduler_hints)
        self.assertNotEqual(nova.get_server_hypervisor(sender),
                            nova.get_server_hypervisor(receiver))


# --- l3-ha extension VM to VM scenario ---------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3haNetworkTest(NetworkTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_fixture(stacks.L3haPeerServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3haSameHostNetworkTest(SameHostNetworkTest):
    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.L3haSameHostServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
@nova.skip_if_missing_hypervisors(count=2, state='up', status='enabled')
class L3haDifferentHostNetworkTest(DifferentHostNetworkTest):
    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.L3haDifferentHostServerStackFixture)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import pytest
import testtools

import tobiko
from tobiko.openstack import nova
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko import podified
from tobiko.shell import ip
from tobiko.shell import ping
from tobiko.shell import sh


LOG = log.getLogger(__name__)


@pytest.mark.minimal
class OpenstackNodesTest(testtools.TestCase):

    topology = tobiko.required_fixture(
        topology.get_default_openstack_topology_class())

    def test_public_ips(self):
        ips = dict()
        for node in self.topology.nodes:
            ping.ping(node.public_ip).assert_replied()
            other = ips.setdefault(node.public_ip, node)
            if node is not other:
                tobiko.fail(f"Nodes {node.name} and {other.name} have the "
                            f"same IP: {node.public_ip}")

    def test_hostnames(self):
        hostnames = dict()
        for node in self.topology.nodes:
            if node.ssh_client is None:
                LOG.debug(f'Node {node.hostname} has no ssh_client')
                continue
            hostname = sh.get_hostname(ssh_client=node.ssh_client)
            self.assertTrue(hostname.startswith(node.name))
            other = hostnames.setdefault(hostname, node)
            if node is not other:
                tobiko.fail(f"Nodes {node.name} and {other.name} have the "
                            f"same hostname: {hostname}")

    def test_network_namespaces(self):
        for node in self.topology.nodes:
            namespaces_ips = {}
            namespaces = ip.list_network_namespaces(ssh_client=node.ssh_client)
            for namespace in namespaces:
                ips = ip.list_ip_addresses(ssh_client=node.ssh_client,
                                           network_namespace=namespace)
                other_ips = namespaces_ips.setdefault(namespace, ips)
                if ips is not other_ips:
                    tobiko.fail(f"Duplicate network namespace {namespace} in "
                                f"node {node.name}: {other_ips}, {ips}")

    @podified.skip_if_not_podified
    def test_systemd_services(self):
        """Verify expected systemd services are running on compute nodes"""
        # Define the services to check on compute nodes
        # NOTE(ralonsoh): remove the OVN metadata agent once it is removed
        # (currently it is only deprecated).
        services_to_check = [
            nova.ISCSID,
            nova.NOVA_COMPUTE,
            neutron.OVN_CONTROLLER,
            (neutron.NEUTRON_OVN_METADATA_AGENT, neutron.NEUTRON_OVN_AGENT),
        ]

        # Get compute nodes
        compute_nodes = topology.list_openstack_nodes(
            group=podified.ALL_COMPUTES_GROUP_NAME)

        for node in compute_nodes:
            LOG.info(f'Checking systemd services on node {node.name}')
            for service_itr in services_to_check:
                if not tobiko.is_collection(service_itr):
                    service_itr = tuple([service_itr])

                service_names = []
                for service_key in service_itr:
                    service_names.append(self.topology.get_agent_service_name(
                        service_key))

                LOG.debug(f'Checking service(s) {service_names} on node '
                          f'{node.name}')
                service_units = [sname + '.service' for sname in service_names]
                running_units = sh.wait_for_active_systemd_units(
                    *service_units,
                    ssh_client=node.ssh_client,
                    sudo=True,
                    timeout=30)
                LOG.info(f'Service {running_units.first} is active on node '
                         f'{node.name}')



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
from collections import abc
import configparser
import functools
import re
import typing
from urllib import parse
import weakref

import netaddr
from oslo_log import log
from packaging import version

import tobiko
from tobiko.shell import dns_ping
from tobiko.shell import dhcp_ping
from tobiko.shell import files
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import keystone
from tobiko.openstack.topology import _address
from tobiko.openstack.topology import _config
from tobiko.openstack.topology import _connection
from tobiko.openstack.topology import _exception


LOG = log.getLogger(__name__)


PatternType = type(re.compile(r'.*'))
OpenstackGroupNameType = typing.Union[str, typing.Pattern]
OpenstackGroupNamesType = typing.Optional[typing.Union[
    OpenstackGroupNameType, typing.Iterable[OpenstackGroupNameType]]]


HostAddressType = typing.Union[str, netaddr.IPAddress]


def list_openstack_nodes(addresses: typing.Iterable[netaddr.IPAddress] = None,
                         group: OpenstackGroupNamesType = None,
                         hostnames: typing.Iterable[str] = None,
                         topology: 'OpenStackTopology' = None,
                         **kwargs) \
        -> tobiko.Selection['OpenStackTopologyNode']:
    if topology is None:
        topology = get_openstack_topology()

    nodes: tobiko.Selection[OpenStackTopologyNode]
    if group is None:
        nodes = topology.nodes
    elif isinstance(group, str):
        nodes = topology.get_group(group=group)
    elif isinstance(group, PatternType):
        nodes = topology.get_groups(groups=[group])
    else:
        assert isinstance(group, abc.Iterable)
        nodes = topology.get_groups(groups=group)

    return select_openstack_nodes(nodes,
                                  addresses=addresses,
                                  hostnames=hostnames,
                                  **kwargs)


def split_addresses_and_names(*hosts: HostAddressType) \
        -> typing.Tuple[typing.Set[netaddr.IPAddress], typing.Set[str]]:
    addresses = set()
    hostnames = set()
    for host in hosts:
        try:
            addresses.add(netaddr.IPAddress(host))
        except netaddr.AddrFormatError:
            hostnames.add(host)
    return addresses, hostnames


def select_openstack_nodes(
        nodes: tobiko.Selection['OpenStackTopologyNode'],
        addresses: typing.Iterable[netaddr.IPAddress] = None,
        hostnames: typing.Iterable[str] = None,
        **kwargs) \
        -> tobiko.Selection['OpenStackTopologyNode']:
    for selector in addresses, hostnames:
        if selector is not None and not selector:
            return tobiko.Selection()

    if addresses is not None:
        _addresses = set(addresses)
        nodes = nodes.select(
            lambda node: bool(set(node.addresses) & _addresses))

    if hostnames is not None:
        names = {node_name_from_hostname(hostname)
                 for hostname in hostnames}
        nodes = nodes.select(lambda node: node.name in names)

    if kwargs:
        nodes = nodes.with_attributes(**kwargs)
    return nodes


def find_openstack_node(addresses: typing.Iterable[netaddr.IPAddress] = None,
                        group: OpenstackGroupNamesType = None,
                        hostnames: typing.Iterable[str] = None,
                        topology: 'OpenStackTopology' = None,
                        unique=False,
                        **kwargs) \
        -> 'OpenStackTopologyNode':
    nodes = list_openstack_nodes(topology=topology,
                                 addresses=addresses,
                                 group=group,
                                 hostnames=hostnames,
                                 **kwargs)
    if unique:
        return nodes.unique
    else:
        return nodes.first


def get_l3_agent_mode(name: str = None,
                      hostname: str = None,
                      address=None,
                      topology: 'OpenStackTopology' = None):
    return get_openstack_node(name=name,
                              hostname=hostname,
                              address=address,
                              topology=topology).l3_agent_mode


def get_openstack_node(name: str = None,
                       hostname: str = None,
                       address=None,
                       topology: 'OpenStackTopology' = None) \
        -> 'OpenStackTopologyNode':
    if topology is None:
        topology = get_openstack_topology()
    return topology.get_node(name=name,
                             hostname=hostname,
                             address=address)


def list_openstack_node_groups(topology=None):
    topology = topology or get_openstack_topology()
    return topology.groups


def get_default_openstack_topology_class() -> typing.Type:
    return DEFAULT_TOPOLOGY_CLASS


def set_default_openstack_topology_class(topology_class: typing.Type):
    # pylint: disable=global-statement
    if not issubclass(topology_class, OpenStackTopology):
        raise TypeError(f"'{topology_class}' is not subclass of "
                        f"'{OpenStackTopology}'")
    global DEFAULT_TOPOLOGY_CLASS
    DEFAULT_TOPOLOGY_CLASS = topology_class


def get_agent_service_name(agent_name: str) -> str:
    topology_class = get_default_openstack_topology_class()
    return topology_class.get_agent_service_name(agent_name)


def get_agent_container_name(agent_name: str) -> str:
    topology_class = get_default_openstack_topology_class()
    return topology_class.get_agent_container_name(agent_name)


def check_systemd_monitors_agent(hostname: str, agent_name: str) -> bool:
    ssh_client = get_openstack_node(hostname=hostname).ssh_client
    services = sh.execute('systemctl list-units --type=service --all '
                          '--no-pager --no-legend',
                          ssh_client=ssh_client,
                          sudo=True).stdout.split('\n')
    for service in services:
        if get_agent_service_name(agent_name) in service:
            return True
    return False


class UnknowOpenStackServiceNameError(tobiko.TobikoException):
    message = ("Unknown service name for agent name '{agent_name}' and "
               "topology class '{topology_class}'")


class UnknowOpenStackContainerNameError(tobiko.TobikoException):
    message = ("Unknown container name for agent name '{agent_name}' and "
               "topology class '{topology_class}'")


class UnknownOpenStackConfigurationFile(tobiko.TobikoException):
    message = ("Unknown configuration file '{file_name}'")


class OpenStackTopologyNode:

    _podman_client = None

    def __init__(self,
                 topology: 'OpenStackTopology',
                 name: str,
                 ssh_client: typing.Optional[ssh.SSHClientFixture],
                 addresses: typing.Iterable[netaddr.IPAddress],
                 hostname: str):
        self._topology = weakref.ref(topology)
        self.name = name
        self.ssh_client = ssh_client
        self.groups: typing.Set[str] = set()
        self.addresses: tobiko.Selection[netaddr.IPAddress] = tobiko.select(
            addresses)
        self.hostname: str = hostname

    _connection: typing.Optional[sh.ShellConnection] = None

    @property
    def connection(self) -> sh.ShellConnection:
        if self._connection is None:
            self._connection = sh.shell_connection(self.ssh_client)
        return self._connection

    @property
    def topology(self):
        return self._topology()

    def add_group(self, group: str):
        self.groups.add(group)

    @property
    def public_ip(self):
        return self.addresses[0]

    @property
    def ssh_parameters(self):
        return self.ssh_client.setup_connect_parameters()

    @property
    def podman_client(self):
        podman_client = self._podman_client
        if not podman_client:
            from tobiko import podman
            self._podman_client = podman_client = podman.get_podman_client(
                ssh_client=self.ssh_client)
        return podman_client

    l3_agent_conf_path = '/etc/neutron/l3_agent.ini'
    _l3_agent_mode: typing.Optional[str] = None

    @property
    def l3_agent_mode(self) -> typing.Optional[str]:
        if self._l3_agent_mode is None:
            try:
                self._l3_agent_mode = neutron.get_l3_agent_mode(
                    l3_agent_conf_path=self.l3_agent_conf_path,
                    connection=self.connection)
            except sh.ShellCommandFailed:
                LOG.debug('Unable to read L3 agent mode for host '
                          f'{self.hostname}. Assuming legacy mode.',
                          exc_info=1)
                self._l3_agent_mode = 'legacy'
        return self._l3_agent_mode

    def __repr__(self):
        return "{cls!s}<name={name!r}>".format(cls=type(self).__name__,
                                               name=self.name)


class OpenStackTopology(tobiko.SharedFixture):

    config = tobiko.required_fixture(_config.OpenStackTopologyConfig)

    agent_to_service_name_mappings = {
        neutron.DHCP_AGENT: 'devstack@q-dhcp',
        neutron.L3_AGENT: 'devstack@q-l3',
        neutron.OPENVSWITCH_AGENT: 'devstack@q-agt',
        neutron.METADATA_AGENT: 'devstack@q-meta',
        neutron.OVN_METADATA_AGENT: 'devstack@q-ovn-metadata-agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'devstack@q-ovn-metadata-agent',
        neutron.NEUTRON_OVN_AGENT: 'devstack@q-ovn-agent',
        neutron.OVN_CONTROLLER: 'ovn-controller',
        neutron.OVN_BGP_AGENT: 'ovn-bgp-agent',
        neutron.FRR: 'frr',
        neutron.NEUTRON: 'neutron'
    }
    agent_to_container_name_mappings: typing.Dict[str, str] = {}

    has_containers = False

    config_file_mappings = {
        'ml2_conf.ini': '/etc/neutron/plugins/ml2/ml2_conf.ini',
        'bgp-agent.conf': '/etc/ovn-bgp-agent/bgp-agent.conf',
        'octavia.conf': '/etc/octavia/octavia.conf'
    }

    _connections = tobiko.required_fixture(
        _connection.SSHConnectionManager)

    # In Devstack based env logs can be accessed by journalctl
    file_digger_class: typing.Type[files.LogFileDigger] = \
        files.JournalLogDigger

    log_names_mappings: dict = {}

    def __init__(self):
        super(OpenStackTopology, self).__init__()
        self._names: typing.Dict[str, OpenStackTopologyNode] = (
            collections.OrderedDict())
        self._groups: typing.Dict[str, tobiko.Selection] = (
            collections.OrderedDict())
        self._addresses: typing.Dict[netaddr.IPAddress,
                                     OpenStackTopologyNode] = (
            collections.OrderedDict())
        # This is dict which handles mapping of the log file and systemd_unit
        # (if needed) for the OpenStack services.
        # In case of Devstack topology file name in fact name of the systemd
        # unit as logs are stored in journalctl
        if not self.log_names_mappings:
            self.log_names_mappings = {
                neutron.SERVER: (
                    'devstack@' +
                    tobiko.tobiko_config().topology.devstack_neutron_service)
            }

    def setup_fixture(self):
        self.discover_nodes()

    def cleanup_fixture(self):
        tobiko.cleanup_fixture(self._connections)
        self._names.clear()
        self._groups.clear()
        self._addresses.clear()

    @classmethod
    def get_agent_service_name(cls, agent_name: str) -> str:
        try:
            return cls.agent_to_service_name_mappings[agent_name]
        except KeyError:
            pass
        raise UnknowOpenStackServiceNameError(agent_name=agent_name,
                                              topology_class=cls)

    @classmethod
    def get_agent_container_name(cls, agent_name: str) -> str:
        try:
            return cls.agent_to_container_name_mappings[agent_name]
        except KeyError:
            pass
        raise UnknowOpenStackContainerNameError(agent_name=agent_name,
                                                topology_class=cls)

    def get_config_file_path(self, file_name: str) -> str:
        try:
            return self.config_file_mappings[file_name]
        except KeyError as e:
            raise UnknownOpenStackConfigurationFile(file_name=file_name) from e

    def get_log_file_digger(self,
                            service_name: str,
                            pattern: typing.Optional[str] = None,
                            groups: typing.Optional[typing.List[str]] = None,
                            sudo=True,
                            **execute_params) -> \
            files.MultihostLogFileDigger:
        digger = files.MultihostLogFileDigger(
            filename=self.log_names_mappings[service_name],
            pattern=pattern,
            file_digger_class=self.file_digger_class,
            sudo=sudo,
            **execute_params)
        if groups is None:
            nodes = self.nodes
        else:
            nodes = self.get_groups(groups=groups)
        for node in nodes:
            digger.add_host(hostname=node.name,
                            ssh_client=node.ssh_client)
        return digger

    def assert_containers_running(self, expected_containers,
                                  group=None,
                                  full_name=True, bool_check=False,
                                  nodenames=None):
        pass

    def list_containers_td(self, group=None):
        pass

    def run_octavia_ovn_db_sync(
            self,
            ssh_client: ssh.SSHClientFixture) -> sh.ShellExecuteResult:
        """Run Octavia OVN database synchronization tool

        In DevStack/OpenstackTopology, the tool needs to be run from the
        ovn-octavia-provider directory with a properly configured venv:
        1. Create/use venv
        2. Install package in editable mode
        3. Run octavia-ovn-db-sync-util

        Args:
            ssh_client: SSH client to the controller node

        Returns:
            Shell execution result with stdout/stderr
        """
        LOG.info("Running Octavia OVN DB sync tool in DevStack environment")

        ovn_octavia_path = '/opt/stack/ovn-octavia-provider'

        # Get the owner of the directory to run commands as that user
        owner_result = sh.execute(
            f'stat -c %U {ovn_octavia_path}',
            ssh_client=ssh_client
        )
        dir_owner = owner_result.stdout.strip()
        LOG.info(f"Directory owner: {dir_owner}")

        # Determine if we need to use sudo
        current_user_result = sh.execute('whoami', ssh_client=ssh_client)
        current_user = current_user_result.stdout.strip()
        LOG.info(f"Current user: {current_user}")

        # Prepare command prefix based on user
        if current_user != dir_owner:
            cmd_prefix = f'sudo -u {dir_owner} bash -c "'
            cmd_suffix = '"'
            LOG.info(f"Will run commands as {dir_owner} using sudo")
        else:
            cmd_prefix = ''
            cmd_suffix = ''
            LOG.info(f"Running commands as current user {current_user}")

        # Check if venv exists, create if not
        check_venv_cmd = (
            f'{cmd_prefix}test -d {ovn_octavia_path}/venv && '
            f'echo env_exists || echo env_not_exists{cmd_suffix}'
        )
        check_venv = sh.execute(check_venv_cmd, ssh_client=ssh_client)

        if 'env_exists' not in check_venv.stdout:
            LOG.info("Creating venv for ovn-octavia-provider")
            cmd_create_venv = (
                f'{cmd_prefix}cd {ovn_octavia_path} && '
                f'python3 -m venv venv && '
                f'source venv/bin/activate && '
                f'pip install -e .{cmd_suffix}'
            )
            sh.execute(cmd_create_venv, ssh_client=ssh_client, timeout=300)
            LOG.info("Venv created and package installed")
        else:
            LOG.info("Venv already exists")

        # Run the sync tool
        cmd = (f'{cmd_prefix}source {ovn_octavia_path}/venv/bin/activate && '
               f'octavia-ovn-db-sync-util{cmd_suffix}')

        LOG.info(f"Executing: {cmd}")
        output = sh.execute(cmd, ssh_client=ssh_client, timeout=300)
        LOG.info(f"Sync tool output: {output.stdout}")
        return output

    def discover_nodes(self):
        self.discover_ssh_proxy_jump_node()
        self.discover_configured_nodes()
        if keystone.has_keystone_credentials():
            self.discover_controller_nodes()
            self.discover_compute_nodes()

    def discover_ssh_proxy_jump_node(self):
        ssh_client = ssh.ssh_proxy_client()
        if ssh_client is not None:
            self.add_node(ssh_client=ssh_client,
                          group='proxy_jump')

    def discover_configured_nodes(self):
        for address in self.config.conf.nodes or []:
            self.add_node(address=address)

    def discover_controller_nodes(self):
        endpoints = keystone.list_endpoints(interface='public')
        addresses = set(parse.urlparse(endpoint.url).hostname
                        for endpoint in endpoints)
        for address in addresses:
            try:
                self.add_node(address=address, group='controller')
            except _connection.UreachableSSHServer as ex:
                LOG.debug(f"Unable to SSH to end point address '{address}'. "
                          f"{ex}")

    def discover_compute_nodes(self):
        for hypervisor in nova.list_hypervisors():
            self.add_node(hostname=hypervisor.hypervisor_hostname,
                          address=hypervisor.host_ip,
                          group='compute')

    def add_node(self,
                 hostname: typing.Optional[str] = None,
                 address: typing.Optional[str] = None,
                 group: typing.Optional[str] = None,
                 ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
                 **create_params) \
            -> OpenStackTopologyNode:
        if ssh_client is not None:
            # detect all global addresses from remote server
            try:
                hostname = sh.get_hostname(ssh_client=ssh_client)
            except Exception:
                LOG.exception("Unable to get node hostname from "
                              f"{ssh_client}")
                ssh_client = None
        name = hostname and node_name_from_hostname(hostname) or None

        addresses: typing.List[netaddr.IPAddress] = []
        if address:
            # add manually configure addresses first
            addresses.extend(self._list_addresses(address))
        addresses = tobiko.select(remove_duplications(addresses))

        try:
            node = self.get_node(name=name, address=addresses)
        except _exception.NoSuchOpenStackTopologyNode:
            node = self._add_node(addresses=addresses,
                                  hostname=hostname,
                                  ssh_client=ssh_client,
                                  **create_params)

        if group:
            # Add group anyway even if the node hasn't been added
            group_nodes = self.add_group(group=group)
            if node and node not in group_nodes:
                group_nodes.append(node)
                node.add_group(group=group)

        return node

    def _add_node(self,
                  addresses: typing.List[netaddr.IPAddress],
                  hostname: str = None,
                  ssh_client: ssh.SSHClientFixture = None,
                  create_ssh_client: bool = True,
                  **create_params):
        if ssh_client is None and create_ssh_client:
            ssh_client = self._ssh_connect(hostname=hostname,
                                           addresses=addresses)
        addresses.extend(self._list_addresses_from_host(ssh_client=ssh_client))
        addresses = tobiko.select(remove_duplications(addresses))
        hostname = hostname or sh.get_hostname(ssh_client=ssh_client)
        name = node_name_from_hostname(hostname)
        try:
            node = self._names[name]
        except KeyError:
            ssh_login = (
                ssh_client.login if ssh_client else "No SSH Client configured")
            LOG.debug("Add topology node:\n"
                      f" - name: {name}\n"
                      f" - hostname: {hostname}\n"
                      f" - login: {ssh_login}\n"
                      f" - addresses: {addresses}\n")
            self._names[name] = node = self.create_node(name=name,
                                                        hostname=hostname,
                                                        ssh_client=ssh_client,
                                                        addresses=addresses,
                                                        **create_params)

        for address in addresses:
            address_node = self._addresses.setdefault(address, node)
            if address_node is not node:
                LOG.warning(f"Address '{address}' of node '{name}' is already "
                            f"used by node '{address_node.name}'")
        return node

    def get_node(self,
                 name: str = None,
                 hostname: str = None,
                 address=None) \
            -> 'OpenStackTopologyNode':
        if name is None and hostname is not None:
            name = node_name_from_hostname(hostname)
        details = {}
        if name is not None:
            tobiko.check_valid_type(name, str)
            details['name'] = name
            try:
                return self._names[name]
            except KeyError:
                pass
        if address is not None:
            details['address'] = address
            for address in self._list_addresses(address):
                try:
                    return self._addresses[address]
                except KeyError:
                    pass
        raise _exception.NoSuchOpenStackTopologyNode(details=details)

    def create_node(self, name, ssh_client, **kwargs):
        return OpenStackTopologyNode(topology=self, name=name,
                                     ssh_client=ssh_client, **kwargs)

    @property
    def nodes(self) -> tobiko.Selection[OpenStackTopologyNode]:
        return tobiko.select(self.get_node(name)
                             for name in self._names)

    def add_group(self, group: str) -> tobiko.Selection:
        try:
            return self._groups[group]
        except KeyError:
            self._groups[group] = nodes = self.create_group()
            return nodes

    @staticmethod
    def create_group() -> tobiko.Selection[OpenStackTopologyNode]:
        return tobiko.Selection()

    def get_group(self, group: str) \
            -> tobiko.Selection[OpenStackTopologyNode]:
        tobiko.check_valid_type(group, str)
        try:
            return tobiko.Selection(self._groups[group])
        except KeyError as ex:
            raise _exception.NoSuchOpenStackTopologyNodeGroup(
                group=group) from ex

    def list_group_names(self,
                         *matchers: 'MatchStringType') \
            -> typing.List[str]:
        group_names: typing.List[str] = list(self._groups.keys())
        if matchers and group_names:
            group_names = match_strings(group_names, *matchers)
        return group_names

    def get_groups(self, groups: typing.Iterable['MatchStringType']) -> \
            tobiko.Selection[OpenStackTopologyNode]:
        node_names: typing.Set[str] = set()
        for group in self.list_group_names(*groups):
            node_names.update(node.name for node in self.get_group(group))
        return self.nodes.select(lambda node: node.name in node_names)

    @property
    def groups(self) -> typing.List[str]:
        return list(self._groups)

    def _ssh_connect(self,
                     addresses: typing.List[netaddr.IPAddress],
                     hostname: str = None,
                     **connect_params) -> ssh.SSHClientFixture:
        if hostname is not None:
            try:
                return tobiko.setup_fixture(
                    ssh.ssh_client(hostname, **connect_params))
            except Exception:
                LOG.debug(f'Unable to connect to {hostname} using regular '
                          'SSH configuration')

        try:
            return _connection.ssh_connect(addresses,
                                           **connect_params)
        except _connection.UreachableSSHServer:
            for proxy_node in self.nodes:
                proxy_client = proxy_node.ssh_client
                if proxy_client:
                    LOG.debug("Try connecting through a proxy node "
                              f"'{proxy_node.name}'")
                    try:
                        return self._ssh_connect_with_proxy_client(
                            addresses, proxy_client, **connect_params)
                    except _connection.UreachableSSHServer:
                        pass
            raise

    def _ssh_connect_with_proxy_client(self, addresses, proxy_client,
                                       **connect_params) -> \
            ssh.SSHClientFixture:
        ssh_client = _connection.ssh_connect(addresses,
                                             proxy_client=proxy_client,
                                             **connect_params)
        addresses = self._list_addresses_from_host(ssh_client=ssh_client)
        try:
            LOG.debug("Try connecting through an address that doesn't require "
                      "an SSH proxy host")
            return _connection.ssh_connect(addresses, **connect_params)
        except _connection.UreachableSSHServer:
            return ssh_client

    @property
    def ip_version(self) -> typing.Optional[int]:
        ip_version = self.config.conf.ip_version
        return ip_version and int(ip_version) or None

    def _list_addresses_from_host(
            self,
            ssh_client: typing.Optional[ssh.SSHClientFixture]):
        if not ssh_client:
            return tobiko.Selection()
        return ip.list_ip_addresses(ssh_client=ssh_client,
                                    ip_version=self.ip_version,
                                    scope='global')

    def _list_addresses(self, obj) -> typing.List[netaddr.IPAddress]:
        return _address.list_addresses(obj,
                                       ip_version=self.ip_version,
                                       ssh_config=True)

    def check_or_start_background_vm_ping(self, server_ip, ssh_client=None):  # noqa; pylint: disable=W0613
        tobiko.skip_test("Background ping not supported by "
                         "this topology class.")

    def check_or_start_background_iperf_connection(
            self,
            server_ip: typing.Union[str, netaddr.IPAddress],  # noqa; pylint: disable=W0613
            port: int,  # noqa; pylint: disable=W0613
            protocol: str,  # noqa; pylint: disable=W0613
            ssh_client: ssh.SSHClientType = None,  # noqa; pylint: disable=W0613
            iperf3_server_ssh_client: ssh.SSHClientType = None):  # noqa; pylint: disable=W0613
        tobiko.skip_test("Background iperf not supported by "
                         "this topology class.")

    def check_or_start_background_http_ping(
            self,
            server_ip: typing.Union[str, netaddr.IPAddress],  # noqa; pylint: disable=W0613
            ssh_client: ssh.SSHClientType = None):  # noqa; pylint: disable=W0613
        tobiko.skip_test("Background HTTP ping test is not supported by "
                         "this topology class.")

    def check_or_start_background_dhcp_ping(
            self,
            ssh_client: ssh.SSHClientType):
        sh.check_or_start_external_process(
            start_function=dhcp_ping.start_dhcp_ping_process,
            stop_function=dhcp_ping.stop_dhcp_ping_process,
            liveness_function=dhcp_ping.dhcp_ping_process_alive,
            check_function=dhcp_ping.check_dhcp_ping_results,
            ssh_client=ssh_client)

    def check_or_start_background_dns_ping(
            self,
            ip_address: typing.Union[str, netaddr.IPAddress],
            fqdn: str,
            ssh_client: ssh.SSHClientType):
        sh.check_or_start_external_process(
            start_function=dns_ping.start_dns_ping_process,
            stop_function=dns_ping.stop_dns_ping_process,
            liveness_function=dns_ping.dns_ping_process_alive,
            check_function=dns_ping.check_dns_ping_results,
            ip_address=ip_address,
            fqdn=fqdn,
            ssh_client=ssh_client)


def get_openstack_topology(topology_class: typing.Type = None) -> \
        OpenStackTopology:
    if topology_class:
        if not issubclass(topology_class, OpenStackTopology):
            raise TypeError(f"'{topology_class}' is not subclass of "
                            f"'{OpenStackTopology}'")
    else:
        topology_class = get_default_openstack_topology_class()
    return tobiko.setup_fixture(topology_class)


def get_log_file_digger(
        service_name: str,
        pattern: str = None,
        groups: typing.List[str] = None,
        topology: OpenStackTopology = None,
        sudo=True,
        **execute_params) \
        -> files.MultihostLogFileDigger:
    if topology is None:
        topology = get_openstack_topology()
    return topology.get_log_file_digger(service_name=service_name,
                                        pattern=pattern,
                                        groups=groups,
                                        sudo=sudo,
                                        **execute_params)


def get_config_file_path(file_name: str) -> str:
    topology = get_openstack_topology()
    return topology.get_config_file_path(file_name)


def get_config_setting(file_name: str,
                       ssh_client: ssh.SSHClientFixture,
                       param: str,
                       section: str = None) -> typing.Optional[str]:
    config_file_path = get_config_file_path(file_name)
    config_file_content = sh.execute(f'cat {config_file_path}',
                                     ssh_client=ssh_client, sudo=True).stdout
    config = configparser.ConfigParser()
    config.read_string(config_file_content)
    if section is None:
        value = config.defaults().get(param)
    elif section not in config.sections():
        value = None
    else:
        value = config[section].get(param)

    return value


def get_rhosp_version():
    ssh_client = list_openstack_nodes(group='controller')[0].ssh_client
    rhosp_release = sh.execute('cat /etc/rhosp-release',
                               ssh_client=ssh_client).stdout
    rhosp_version = re.search(r"[0-9]*\.[0-9]*\.[0-9]*", rhosp_release)[0]
    return rhosp_version


def get_nova_version_from_container():
    ssh_client = list_openstack_nodes(group='controller')[0].ssh_client
    container_runtime_cmd = 'podman'
    try:
        cmd = (container_runtime_cmd +
               ' exec -uroot nova_conductor nova-manage --version')
        return sh.execute(cmd,
                          ssh_client=ssh_client,
                          sudo=True).stdout
    except sh.ShellCommandFailed:
        pass


# During a execution of tobiko, openstack version does not change, so let's
# cache the output of this function
@functools.lru_cache()
def get_openstack_version():
    try:
        return get_rhosp_version()
    except (TypeError, sh.ShellCommandFailed):
        pass

    nova_version = get_nova_version_from_container()
    if nova_version is None:
        ssh_client = list_openstack_nodes(group='controller')[0].ssh_client
        nova_version = sh.execute('nova-manage --version',
                                  ssh_client=ssh_client,
                                  sudo=True).stdout
    os_to_nova_versions = {'13.0.0': '17',  # Queens
                           '16.0.0': '19',  # Stein
                           '16.1.0': '20',  # Train
                           '17.0.0': '23'}  # Wallaby
    for os_version, nova_major_version in os_to_nova_versions.items():
        if nova_version.split('.')[0] == nova_major_version:
            return os_version


DEFAULT_TOPOLOGY_CLASS = OpenStackTopology


def node_name_from_hostname(hostname: str):
    return hostname.split('.', 1)[0].lower()


def remove_duplications(items: typing.List) -> typing.List:
    # use all items as dictionary keys to remove duplications
    mapping = collections.OrderedDict((k, None) for k in items)
    return list(mapping.keys())


def verify_osp_version(required_version, higher=False, lower=False):
    try:
        current_version = get_openstack_version()
    except Exception:
        current_version = None
    if current_version is None:
        return True

    required_version_parsed = version.parse(required_version)
    current_version_parsed = version.parse(current_version)

    if higher and lower:
        raise RuntimeError
    elif not higher and not lower:  # this means equal
        return current_version_parsed == required_version_parsed
    elif higher:
        return current_version_parsed > required_version_parsed
    elif lower:
        return current_version_parsed < required_version_parsed


def skip_unless_osp_version(required_version, higher=False, lower=False):
    skip_msg = "OSP version doesn't match the requirement"
    return tobiko.skip_unless(skip_msg, verify_osp_version,
                              required_version, higher, lower)


MatchStringType = typing.Union[str, typing.Pattern]


def match_strings(strings: typing.Iterable[str],
                  *matchers: MatchStringType) -> \
        typing.List[str]:
    matching: typing.List[str] = []
    for matcher in matchers:
        tobiko.check_valid_type(matcher, str, PatternType)
        if isinstance(matcher, str):
            if matcher in strings:
                matching.append(matcher)
        else:
            assert isinstance(matcher, PatternType)
            for string in strings:
                if matcher.match(string):
                    matching.append(string)
    return matching



# Copyright 2025 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import csv
import io
import re
import typing

from oslo_log import log


LOG = log.getLogger(__name__)

TableDataType = typing.Iterable[typing.Union[typing.Dict[str, typing.Any],
                                             tuple]]
OptionalTableDataType = typing.Optional[TableDataType]
TableSetItemValueType = typing.Union[typing.Dict[str, typing.Any],
                                     TableDataType]


class TableData(collections.UserList):
    """
    A list-like class for managing tabular data where each item is a dictionary
    and all dictionaries are guaranteed to have the same set of keys (schema).

    This class enforces the schema upon initialization and when new items are
    added. It provides pandas-like functionality for data manipulation.
    """
    _schema: typing.List[str] = []  # Stores the ordered list of expected keys
    data: typing.List[typing.Dict[str, typing.Any]] = []

    def __init__(self,
                 initial_data: OptionalTableDataType = None,
                 columns: typing.List[str] = None):
        """
        Initializes the TableData object.

        Args:
            initial_data: An optional iterable of dictionaries to pre-populate
                          the table.
                          If provided, the schema will be inferred from the
                          first dictionary (if any), and all subsequent
                          dictionaries must conform to this schema.
                          Optionally, a list of tuples can be provided instead
                          of a list of dictionaries.
        """
        super().__init__()
        self._schema = []  # No schema unless initial_data is provided

        if initial_data:
            # Convert to list to iterate multiple times if needed
            initial_data_list = list(initial_data)

            if columns:
                self._schema = columns
                for item in initial_data_list:
                    if not isinstance(item, tuple):
                        raise ValueError("If columns are provided, all items "
                                         "must be tuples")
                    new_item = dict(zip(columns, item))
                    self._validate_item(new_item)
                    self.data.append(new_item)
            else:
                for i, item in enumerate(initial_data_list):
                    if not isinstance(item, dict):
                        raise ValueError("If columns are not provided, all "
                                         "items must be dictionaries")
                    if i == 0:
                        # Infer schema from the first item's keys
                        # (order matters)
                        self._schema = list(item.keys())
                    else:
                        self._validate_item(item)
                    self.data.append(item)  # Add to the internal list

    @classmethod
    def from_dict(cls, data_dict: typing.Dict[str, typing.List[typing.Any]]):
        """
        Create a TableData object from a dictionary of lists (pandas-style).

        Args:
            data_dict: Dictionary where keys are column names and values are
                      lists of values for that column.

        Returns:
            TableData object
        """
        if not data_dict:
            return cls()

        # Get all keys (column names)
        columns = list(data_dict.keys())

        # Check that all lists have the same length
        lengths = [len(data_dict[col]) for col in columns]
        if not all(length == lengths[0] for length in lengths):
            raise ValueError("All columns must have the same number of rows")

        # Convert to list of dictionaries
        num_rows = lengths[0] if lengths else 0
        rows = []
        for i in range(num_rows):
            row = {col: data_dict[col][i] for col in columns}
            rows.append(row)

        return cls(rows)

    @classmethod
    def read_csv(cls, stream_or_string, header=None, skiprows=0,
                 columns=None, sep=',', delim_whitespace=False):
        """
        Read CSV data into a TableData object.

        Args:
            stream_or_string: String, StringIO, or file-like object
            header: Row number to use as column names, None means no header
            skiprows: Number of rows to skip at the beginning
            columns: List of column names to use
            sep: Delimiter to use
            delim_whitespace: Use whitespace as delimiter

        Returns:
            TableData object
        """
        stream = cls._prepare_stream(stream_or_string)
        lines = cls._read_csv_lines(stream, skiprows, sep, delim_whitespace)

        if not lines:
            return cls()

        column_names = cls._determine_column_names(lines, header, columns)
        rows = cls._convert_lines_to_rows(lines, column_names, header)

        return cls(rows)

    @classmethod
    def _prepare_stream(cls, stream_or_string):
        """Prepare input stream for reading."""
        if isinstance(stream_or_string, str):
            return io.StringIO(stream_or_string)
        return stream_or_string

    @classmethod
    def _read_csv_lines(cls, stream, skiprows, sep, delim_whitespace):
        """Read and parse CSV lines from stream."""
        # Skip initial rows
        for _ in range(skiprows):
            try:
                next(stream)
            except StopIteration:
                break

        # Handle delim_whitespace
        if delim_whitespace:
            return cls._read_whitespace_delimited(stream)
        elif len(sep) > 1:
            # Handle multi-character delimiters
            return cls._read_multi_char_delimited(stream, sep)
        else:
            reader = csv.reader(stream, delimiter=sep)
            return list(reader)

    @classmethod
    def _read_whitespace_delimited(cls, stream):
        """Read whitespace-delimited data."""
        lines = []
        for line in stream:
            line = line.strip()
            if line:
                # Split on any whitespace and filter empty strings
                parts = [part for part in line.split() if part]
                lines.append(parts)
        return lines

    @classmethod
    def _read_multi_char_delimited(cls, stream, sep):
        """Read data with multi-character delimiter."""
        lines = []
        for line in stream:
            line = line.rstrip('\n\r')  # Remove only line endings
            if line:
                parts = line.split(sep)
                lines.append(parts)
        return lines

    @classmethod
    def _determine_column_names(cls, lines, header, columns):
        """Determine column names from header, columns, or generate defaults.
        """
        if header is not None and header < len(lines):
            return lines[header]
        elif columns:
            return columns
        else:
            # Generate default column names
            num_cols = len(lines[0]) if lines else 0
            return [str(i) for i in range(num_cols)]

    @classmethod
    def _convert_lines_to_rows(cls, lines, column_names, header):
        """Convert parsed lines to list of dictionaries."""
        # Determine which lines contain data
        if header is not None and header < len(lines):
            data_lines = lines[header + 1:]
        else:
            data_lines = lines

        # Convert to dictionaries
        rows = []
        for line in data_lines:
            if len(line) >= len(column_names):
                row = {col: val for col, val in zip(column_names, line)}
                rows.append(row)

        return rows

    @property
    def schema(self) -> typing.List[str]:
        """Returns the ordered list of keys (headers) for the table."""
        return self._schema

    @property
    def columns(self) -> typing.List[str]:
        """Alias for schema to match pandas API."""
        return self._schema

    @columns.setter
    def columns(self, new_columns: typing.List[str]):
        """Set column names (pandas-style)."""
        if len(new_columns) != len(self._schema):
            raise ValueError(
                "Number of new columns must match existing schema")

        # Update all rows with new column names
        new_data = []
        for row in self.data:
            new_row = {new_col: row[old_col]
                       for new_col, old_col in zip(new_columns, self._schema)}
            new_data.append(new_row)

        self._schema = list(new_columns)
        self.data = new_data

    @property
    def empty(self) -> bool:
        """Check if table is empty (pandas-style)."""
        return len(self.data) == 0

    def _validate_item(self, item: typing.Union[typing.Dict[str, typing.Any],
                                                tuple]):
        """
        Internal method to validate if a dictionary conforms to the current
        schema.
        Raises ValueError if the item does not conform.
        """
        if not isinstance(item, dict):
            raise TypeError(
                f"Items must be dictionaries, but got {type(item)}")

        if not self._schema:
            # If schema is not yet set, infer it from the first item
            self._schema = list(item.keys())
            if not self._schema:  # Handle empty dictionary as first item
                raise ValueError("Cannot infer schema from an empty "
                                 "dictionary as the first item.")
        else:
            # Check if the keys match the schema
            if set(item.keys()) != set(self._schema):
                raise ValueError("Dictionary keys do not match schema. "
                                 f"Expected {set(self._schema)}, "
                                 f"got {set(item.keys())}")

    def append(self, item: typing.Union[typing.Dict[str, typing.Any], tuple]):
        """Appends a dictionary to the table, enforcing schema."""
        self._validate_item(item)
        super().append(item)

    def insert(self, i: int, item: typing.Dict[str, typing.Any]):
        """Inserts a dictionary into the table, enforcing schema."""
        self._validate_item(item)
        super().insert(i, item)

    # The correct signature for __setitem__ to avoid MyPy errors
    @typing.overload
    def __setitem__(self,
                    key: typing.SupportsIndex,
                    value: typing.Dict[str, typing.Any]): ...

    @typing.overload
    def __setitem__(self,
                    key: slice,
                    value: TableDataType): ...

    def __setitem__(self,
                    key: typing.Union[typing.SupportsIndex, slice],
                    value: TableSetItemValueType):
        """Sets an item or slice, enforcing schema."""
        if isinstance(key, slice):
            if not isinstance(value, list):
                raise TypeError(
                    "Must assign a list of dictionaries to a slice")
            for item in value:
                self._validate_item(item)
        else:
            if not isinstance(value, dict):
                raise TypeError(
                    "Must assign a dictionaries to an index")
            self._validate_item(value)
        super().__setitem__(key, value)

    # You might also want to override __add__ and __iadd__ for concatenation
    def __add__(self, other: TableDataType):
        if not isinstance(other, list):
            other = list(other)
        new_data = self.data + other
        return type(self)(new_data)  # Create a new re-validated instance

    def __iadd__(self, other: TableDataType):
        for item in other:
            self.append(item)  # Use append to ensure validation
        return self

    def query(self, expr: str) -> 'TableData':
        """
        Query the table using a simple expression (pandas-style).

        Supports expressions like:
        - column_name == "value"
        - column_name != "value"
        - column_name.str.contains("substring")

        Args:
            expr: Query expression string

        Returns:
            New TableData with filtered results
        """
        results = []
        LOG.debug('filtering data with expression: %s' % expr)

        for row in self.data:
            # Simple expression evaluation
            # Replace column names with their values for evaluation
            eval_expr = expr

            # Handle string contains operations
            contains_match = re.search(r'(\w+)\.str\.contains\("([^"]+)"\)',
                                       expr)
            if contains_match:
                col_name, search_str = contains_match.groups()
                if col_name in row:
                    val = str(row[col_name])
                    if search_str in val:
                        results.append(row)
                continue

            # Handle equality/inequality operations
            # Sort schema by length (descending) to replace longer names first
            # This prevents substring matches (e.g., 'resource' vs others)
            for col_name in sorted(self._schema, key=len, reverse=True):
                # Use word boundary regex to match only complete column names
                pattern = r'\b' + re.escape(col_name) + r'\b'
                if re.search(pattern, eval_expr):
                    # Replace column name with quoted value for safe eval
                    col_value = row[col_name]
                    if isinstance(col_value, str):
                        # Use lambda to avoid escape sequence interpretation
                        # Bind col_value to avoid cell-var-from-loop warning
                        eval_expr = re.sub(
                            pattern,
                            typing.cast(typing.Callable[[typing.Any], str],
                                        lambda m, val=col_value: f'"{val}"'),
                            eval_expr)
                    else:
                        # Use lambda to avoid escape sequence interpretation
                        # Bind col_value to avoid cell-var-from-loop warning
                        eval_expr = re.sub(
                            pattern,
                            typing.cast(typing.Callable[[typing.Any], str],
                                        lambda m, val=col_value: str(val)),
                            eval_expr)

            LOG.debug('final expression to evaluate: %s' % eval_expr)
            try:
                # Use eval with restricted namespace to support comparisons
                # while maintaining safety by preventing access to builtins
                if eval( # noqa; pylint: disable=eval-used
                        eval_expr, {"__builtins__": {}}, {}):
                    results.append(row)
            except (SyntaxError, NameError, ValueError):
                # If expression can't be evaluated, skip this row
                continue

        return TableData(results)

    def __getitem__(self, key):
        """
        Support pandas-style column access and boolean indexing.
        """
        if isinstance(key, str):
            # Column access - return a ColumnData object
            return ColumnData([row[key] for row in self.data], key)
        elif isinstance(key, list) and all(isinstance(k, bool) for k in key):
            # Boolean indexing
            if len(key) != len(self.data):
                raise ValueError(
                    "Boolean index length doesn't match table length")
            results = [row for row, include in zip(self.data, key) if include]
            return TableData(results)
        else:
            # Regular list indexing
            return super().__getitem__(key)

    def replace(self, to_replace=None, value=None, regex=False, inplace=False):
        """
        Replace values in the table (pandas-style).

        Args:
            to_replace: Value or dict of values to replace
            value: Value to replace with
            regex: Whether to_replace is a regex pattern
            inplace: Whether to modify this object or return a new one
        """
        target = self if inplace else TableData(self.data)

        for i, row in enumerate(target.data):
            new_row = {}
            for col, val in row.items():
                new_val = val

                if isinstance(to_replace, dict):
                    if col in to_replace:
                        pattern = to_replace[col]
                        if regex:
                            new_val = re.sub(pattern, value or '', str(val))
                        else:
                            if val == pattern:
                                new_val = value
                elif to_replace is not None:
                    if regex:
                        new_val = re.sub(to_replace, value or '', str(val))
                    else:
                        if val == to_replace:
                            new_val = value

                new_row[col] = new_val

            target.data[i] = new_row

        if not inplace:
            return target

    def to_csv(self, path_or_buf=None):
        """
        Write table to CSV format (pandas-style).

        Args:
            path_or_buf: File path or buffer to write to
        """
        output = io.StringIO()

        if self._schema and self.data:
            writer = csv.DictWriter(output, fieldnames=self._schema)
            writer.writeheader()
            writer.writerows(self.data)

        csv_string = output.getvalue()

        if path_or_buf is None:
            return csv_string
        elif hasattr(path_or_buf, 'write'):
            path_or_buf.write(csv_string)
        else:
            with open(path_or_buf, 'w', newline='') as f:
                f.write(csv_string)

    def to_string(self):
        """
        Return string representation of table (pandas-style).
        """
        if not self.data:
            return "Empty TableData"

        # Calculate column widths
        col_widths = {}
        for col in self._schema:
            col_widths[col] = len(col)
            for row in self.data:
                col_widths[col] = max(col_widths[col], len(str(row[col])))

        # Build string representation
        lines = []

        # Header
        header_parts = []
        for col in self._schema:
            header_parts.append(col.ljust(col_widths[col]))
        lines.append('  '.join(header_parts))

        # Data rows
        for row in self.data:
            row_parts = []
            for col in self._schema:
                row_parts.append(str(row[col]).ljust(col_widths[col]))
            lines.append('  '.join(row_parts))

        return '\n'.join(lines)

    def concat(self, others: typing.List['TableData']):
        """
        Concatenate multiple TableData objects (pandas-style).

        Args:
            others: List of other TableData objects to concatenate
        Returns:
            New TableData object with concatenated data
        """
        all_data = list(self.data)
        for other in others:
            all_data.extend(other.data)
        return TableData(all_data)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return f"{type(self).__name__}({self.data!r})"


class ColumnData:
    """
    Represents a column of data with pandas-like functionality.
    """

    def __init__(self, data: typing.List[typing.Any], name=None):
        self.data = data
        self.name = name

    @property
    def values(self):
        """Return values as a list-like object with item() method."""
        return ColumnValues(self.data)

    def unique(self):
        """Return unique values."""
        seen = set()
        unique_vals = []
        for val in self.data:
            if val not in seen:
                seen.add(val)
                unique_vals.append(val)
        return unique_vals

    def tolist(self):
        """Convert to list."""
        return list(self.data)

    def count(self):
        """Count non-null values."""
        return len([val for val in self.data if val is not None])

    @property
    def str(self):
        """String operations."""
        return ColumnStrOperations(self.data)


class ColumnValues:
    """
    Wrapper for column values that provides item() method.
    """

    def __init__(self, data: typing.List[typing.Any]):
        self.data = data

    def item(self):
        """Return single value (like pandas Series.values.item())."""
        if len(self.data) == 1:
            return self.data[0]
        elif len(self.data) == 0:
            raise ValueError("Can't call item() on empty data")
        else:
            raise ValueError("Can only call item() on single-element data")


class ColumnStrOperations:
    """
    String operations on column data.
    """

    def __init__(self, data: typing.List[typing.Any]):
        self.data = data

    def contains(self, pattern: str):
        """Check if string contains pattern."""
        return [pattern in str(val) for val in self.data]


# Static function for concatenating TableData objects (pandas-style)
def concat(tables: typing.List[TableData]) -> TableData:
    """
    Concatenate multiple TableData objects.

    Args:
        tables: List of TableData objects to concatenate
    Returns:
        New TableData object with concatenated data
    """
    if not tables:
        return TableData()

    all_data = []
    for table in tables:
        all_data.extend(table.data)

    return TableData(all_data)
